# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# there are two representations of value and mask this module deal with.
#
# "user"
#   (value, mask) or value.  the latter means no mask.
#   value and mask are strings.
#
# "internal"
#   value and mask are on-wire bytes.
#   mask is None if no mask.

import struct

from os_ken.ofproto import ofproto_common
from os_ken.lib.pack_utils import msg_pack_into
from os_ken.lib import type_desc


# 'OFPXXC_EXPERIMENTER' has not corresponding field in the specification.
# This is transparently value for Experimenter class ID for OXM/OXS.
OFPXXC_EXPERIMENTER = 0xffff


def _get_field_info_by_name(oxx, name_to_field, name):
    try:
        f = name_to_field[name]
        t = f.type
        num = f.num
    except KeyError:
        t = type_desc.UnknownType
        if name.startswith('field_'):
            num = int(name.split('_')[1])
        else:
            raise KeyError('unknown %s field: %s' % (oxx.upper(), name))
    return num, t


def _from_user_header(oxx, name_to_field, name):
    (num, t) = _get_field_info_by_name(oxx, name_to_field, name)
    return num


def _from_user(oxx, name_to_field, name, user_value):
    (num, t) = _get_field_info_by_name(oxx, name_to_field, name)
    # the 'list' case below is a bit hack; json.dumps silently maps
    # python tuples into json lists.
    if oxx == 'oxm' and isinstance(user_value, (tuple, list)):
        (value, mask) = user_value
    else:
        value = user_value
        mask = None
    if value is not None:
        value = t.from_user(value)
    if mask is not None:
        mask = t.from_user(mask)
    elif isinstance(value, tuple):
        # This hack is to accomodate CIDR notations with IPv[46]Addr.
        value, mask = value
    return num, value, mask


def _get_field_info_by_number(oxx, num_to_field, n):
    try:
        f = num_to_field[n]
        t = f.type
        name = f.name
    except KeyError:
        t = type_desc.UnknownType
        if isinstance(n, int):
            name = 'field_%d' % (n,)
        else:
            raise KeyError('unknown %s field number: %s' % (oxx.upper(), n))
    return name, t


def _to_user_header(oxx, num_to_field, n):
    (name, t) = _get_field_info_by_number(oxx, num_to_field, n)
    return name


def _to_user(oxx, num_to_field, n, v, m):
    (name, t) = _get_field_info_by_number(oxx, num_to_field, n)
    if v is not None:
        if isinstance(v, (tuple, list)):
            v_len = len(v) * len(v[0])
        else:
            v_len = len(v)
        if hasattr(t, 'size') and t.size != v_len:
            raise Exception(
                'Unexpected %s payload length %d for %s (expected %d)'
                % (oxx.upper(), v_len, name, t.size))
        value = t.to_user(v)
    else:
        value = None
    if m is None:
        user_value = value
    else:
        user_value = (value, t.to_user(m))
    return name, user_value


def _field_desc(num_to_field, n):
    return num_to_field[n]


def _normalize_user(oxx, mod, k, uv):
    try:
        from_user = getattr(mod, oxx + '_from_user')
        (n, v, m) = from_user(k, uv)
    except:
        return (k, uv)
    # apply mask
    if m is not None:
        v = b''.join(struct.Struct(">B").pack(int(x) & int(y)) for (x, y) in zip(v, m))
    try:
        to_user = getattr(mod, oxx + '_to_user')
        (k2, uv2) = to_user(n, v, m)
    except:
        return (k, uv)
    assert k2 == k
    return (k2, uv2)


def _parse_header_impl(mod, buf, offset):
    hdr_pack_str = '!I'
    (header, ) = struct.unpack_from(hdr_pack_str, buf, offset)
    hdr_len = struct.calcsize(hdr_pack_str)
    oxx_type = header >> 9  # class|field
    oxm_hasmask = mod.oxm_tlv_header_extract_hasmask(header)
    oxx_class = oxx_type >> 7
    oxx_length = header & 0xff
    if oxx_class == OFPXXC_EXPERIMENTER:
        # Experimenter OXMs/OXSs have 64-bit header.
        # (vs 32-bit for other OXMs/OXSs)
        exp_hdr_pack_str = '!I'  # experimenter_id
        (exp_id, ) = struct.unpack_from(exp_hdr_pack_str, buf,
                                        offset + hdr_len)
        exp_hdr_len = struct.calcsize(exp_hdr_pack_str)
        assert exp_hdr_len == 4
        oxx_field = oxx_type & 0x7f
        if exp_id == ofproto_common.ONF_EXPERIMENTER_ID and oxx_field == 0:
            # XXX
            # This block implements EXT-256 style experimenter OXM.
            onf_exp_type_pack_str = '!H'
            (exp_type, ) = struct.unpack_from(onf_exp_type_pack_str, buf,
                                              offset + hdr_len + exp_hdr_len)
            exp_hdr_len += struct.calcsize(onf_exp_type_pack_str)
            assert exp_hdr_len == 4 + 2
            num = (exp_id, exp_type)
        else:
            num = (exp_id, oxx_type)
    else:
        num = oxx_type
        exp_hdr_len = 0
    value_len = oxx_length - exp_hdr_len
    if oxm_hasmask:
        value_len //= 2
    assert value_len > 0
    field_len = hdr_len + oxx_length
    total_hdr_len = hdr_len + exp_hdr_len
    return num, total_hdr_len, oxm_hasmask, value_len, field_len


def _parse_header(mod, buf, offset):
    (oxx_type_num, total_hdr_len, hasmask, value_len,
     field_len) = _parse_header_impl(mod, buf, offset)
    return oxx_type_num, field_len - value_len


def _parse(mod, buf, offset):
    (oxx_type_num, total_hdr_len, hasmask, value_len,
     field_len) = _parse_header_impl(mod, buf, offset)
    # Note: OXM/OXS payload length (oxx_len) includes Experimenter ID
    # (exp_hdr_len) for experimenter OXMs/OXSs.
    value_offset = offset + total_hdr_len
    value_pack_str = '!%ds' % value_len
    assert struct.calcsize(value_pack_str) == value_len
    (value, ) = struct.unpack_from(value_pack_str, buf, value_offset)
    if hasmask:
        (mask, ) = struct.unpack_from(value_pack_str, buf,
                                      value_offset + value_len)
    else:
        mask = None
    return oxx_type_num, value, mask, field_len


def _make_exp_hdr(oxx, mod, n):
    exp_hdr = bytearray()
    try:
        get_desc = getattr(mod, '_' + oxx + '_field_desc')
        desc = get_desc(n)
    except KeyError:
        return n, exp_hdr
    if desc._class == OFPXXC_EXPERIMENTER:
        (exp_id, exp_type) = n
        assert desc.experimenter_id == exp_id
        oxx_type = getattr(desc, oxx + '_type')
        if desc.exp_type == 2560:
            # XXX
            # This block implements EXT-256 style experimenter OXM.
            exp_hdr_pack_str = '!IH'  # experimenter_id, exp_type
            msg_pack_into(exp_hdr_pack_str, exp_hdr, 0,
                          desc.experimenter_id, desc.exp_type)
        else:
            assert oxx_type == exp_type | (OFPXXC_EXPERIMENTER << 7)
            exp_hdr_pack_str = '!I'  # experimenter_id
            msg_pack_into(exp_hdr_pack_str, exp_hdr, 0,
                          desc.experimenter_id)
        assert len(exp_hdr) == struct.calcsize(exp_hdr_pack_str)
        n = oxx_type
        assert (n >> 7) == OFPXXC_EXPERIMENTER
    return n, exp_hdr


def _serialize_header(oxx, mod, n, buf, offset):
    try:
        get_desc = getattr(mod, '_' + oxx + '_field_desc')
        desc = get_desc(n)
        value_len = desc.type.size
    except KeyError:
        value_len = 0
    n, exp_hdr = _make_exp_hdr(oxx, mod, n)
    exp_hdr_len = len(exp_hdr)
    pack_str = "!I%ds" % (exp_hdr_len,)
    msg_pack_into(pack_str, buf, offset,
                  (n << 9) | (0 << 8) | (exp_hdr_len + value_len),
                  bytes(exp_hdr))
    return struct.calcsize(pack_str)


def _serialize(oxx, mod, n, value, mask, buf, offset):
    n, exp_hdr = _make_exp_hdr(oxx, mod, n)
    exp_hdr_len = len(exp_hdr)
    value_len = len(value)
    if mask:
        assert value_len == len(mask)
        pack_str = "!I%ds%ds%ds" % (exp_hdr_len, value_len, len(mask))
        msg_pack_into(pack_str, buf, offset,
                      (n << 9) | (1 << 8) | (exp_hdr_len + value_len * 2),
                      bytes(exp_hdr), value, mask)
    else:
        pack_str = "!I%ds%ds" % (exp_hdr_len, value_len,)
        msg_pack_into(pack_str, buf, offset,
                      (n << 9) | (0 << 8) | (exp_hdr_len + value_len),
                      bytes(exp_hdr), value)
    return struct.calcsize(pack_str)



# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import socket
import sys
import unittest
from unittest import mock

from os_ken.lib import pcaplib
from os_ken.lib.packet import packet
from os_ken.lib.packet import zebra
from os_ken.utils import binary_str


PCAP_DATA_DIR = os.path.join(
    os.path.dirname(sys.modules[__name__].__file__),
    '../../packet_data/pcap/')


_patch_frr_v2 = mock.patch(
    'os_ken.lib.packet.zebra._is_frr_version_ge',
    mock.MagicMock(side_effect=lambda x: x == zebra._FRR_VERSION_2_0))


class Test_zebra(unittest.TestCase):
    """
    Test case for os_ken.lib.packet.zebra.
    """

    def _test_pcap_single(self, f):
        zebra_pcap_file = os.path.join(PCAP_DATA_DIR, f + '.pcap')
        # print('*** testing %s' % zebra_pcap_file)

        for _, buf in pcaplib.Reader(open(zebra_pcap_file, 'rb')):
            # Checks if Zebra message can be parsed as expected.
            pkt = packet.Packet(buf)
            zebra_pkts = pkt.get_protocols(zebra.ZebraMessage)
            for zebra_pkt in zebra_pkts:
                self.assertTrue(isinstance(zebra_pkt, zebra.ZebraMessage),
                    'Failed to parse Zebra message: %s' % pkt)
            self.assertTrue(not isinstance(pkt.protocols[-1],
                               (bytes, bytearray)),
                'Some messages could not be parsed in %s: %s' % (f, pkt))

            # Checks if Zebra message can be serialized as expected.
            pkt.serialize()
            self.assertEqual(binary_str(buf), binary_str(pkt.data))

    def test_pcap_quagga(self):
        files = [
            'zebra_v2',
            'zebra_v3',
        ]

        for f in files:
            self._test_pcap_single(f)

    @_patch_frr_v2
    def test_pcap_frr_v2(self):
        files = [
            'zebra_v4_frr_v2',  # API version 4 on FRRouting v2.0
        ]

        for f in files:
            self._test_pcap_single(f)


class TestZebraMessage(unittest.TestCase):

    def test_get_header_size(self):
        self.assertEqual(zebra.ZebraMessage.V0_HEADER_SIZE,
            zebra.ZebraMessage.get_header_size(0))
        self.assertEqual(zebra.ZebraMessage.V1_HEADER_SIZE,
            zebra.ZebraMessage.get_header_size(2))
        self.assertEqual(zebra.ZebraMessage.V3_HEADER_SIZE,
            zebra.ZebraMessage.get_header_size(3))
        self.assertEqual(zebra.ZebraMessage.V3_HEADER_SIZE,
            zebra.ZebraMessage.get_header_size(4))

    def test_get_header_size_invalid_version(self):
        self.assertRaises(ValueError, zebra.ZebraMessage.get_header_size, 0xff)


class TestZebraRedistributeAdd(unittest.TestCase):
    buf = (
        b'\x02'  # route_type
    )
    route_type = zebra.ZEBRA_ROUTE_CONNECT

    def test_parser(self):
        body = zebra.ZebraRedistributeAdd.parse(self.buf, version=3)

        self.assertEqual(self.route_type, body.route_type)

        buf = body.serialize(version=3)

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraIPv4ImportLookup(unittest.TestCase):
    buf = (
        b'\x18'
        b'\xc0\xa8\x01\x01'  # prefix
    )
    prefix = '192.168.1.1/24'
    metric = None
    nexthop_num = 0
    from_zebra = False

    def test_parser(self):
        body = zebra.ZebraIPv4ImportLookup.parse(self.buf)

        self.assertEqual(self.prefix, body.prefix)
        self.assertEqual(self.metric, body.metric)
        self.assertEqual(self.nexthop_num, len(body.nexthops))
        self.assertEqual(self.from_zebra, body.from_zebra)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraIPv4ImportLookupFromZebra(unittest.TestCase):
    buf = (
        b'\xc0\xa8\x01\x01'  # prefix
        b'\x00\x00\x00\x14'  # metric
        b'\x01'              # nexthop_num
        b'\x01'              # nexthop_type
        b'\x00\x00\x00\x02'  # ifindex
    )
    prefix = '192.168.1.1'
    metric = 0x14
    nexthop_num = 1
    nexthop_type = zebra.ZEBRA_NEXTHOP_IFINDEX
    ifindex = 2
    from_zebra = True

    def test_parser(self):
        body = zebra.ZebraIPv4ImportLookup.parse_from_zebra(self.buf)

        self.assertEqual(self.prefix, body.prefix)
        self.assertEqual(self.metric, body.metric)
        self.assertEqual(self.nexthop_num, len(body.nexthops))
        self.assertEqual(self.nexthop_type, body.nexthops[0].type)
        self.assertEqual(self.ifindex, body.nexthops[0].ifindex)
        self.assertEqual(self.from_zebra, body.from_zebra)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraIPv4NexthopLookupMRib(unittest.TestCase):
    buf = (
        b'\xc0\xa8\x01\x01'  # addr
    )
    addr = '192.168.1.1'
    distance = None
    metric = None
    nexthop_num = 0

    def test_parser(self):
        body = zebra.ZebraIPv4NexthopLookupMRib.parse(self.buf)

        self.assertEqual(self.addr, body.addr)
        self.assertEqual(self.distance, body.distance)
        self.assertEqual(self.metric, body.metric)
        self.assertEqual(self.nexthop_num, len(body.nexthops))

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraIPv4NexthopLookupMRibFromZebra(unittest.TestCase):
    buf = (
        b'\xc0\xa8\x01\x01'  # addr
        b'\x01'              # distance
        b'\x00\x00\x00\x14'  # metric
        b'\x01'              # nexthop_num
        b'\x01'              # nexthop_type
        b'\x00\x00\x00\x02'  # ifindex
    )
    addr = '192.168.1.1'
    distance = 1
    metric = 0x14
    nexthop_num = 1
    nexthop_type = zebra.ZEBRA_NEXTHOP_IFINDEX
    ifindex = 2

    def test_parser(self):
        body = zebra.ZebraIPv4NexthopLookupMRib.parse(self.buf)

        self.assertEqual(self.addr, body.addr)
        self.assertEqual(self.distance, body.distance)
        self.assertEqual(self.metric, body.metric)
        self.assertEqual(self.nexthop_num, len(body.nexthops))
        self.assertEqual(self.nexthop_type, body.nexthops[0].type)
        self.assertEqual(self.ifindex, body.nexthops[0].ifindex)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraNexthopUpdateIPv6(unittest.TestCase):
    buf = (
        b'\x00\x0a'          # family
        b'\x40'              # prefix_len
        b'\x20\x01\x0d\xb8'  # prefix
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x14'  # metric
        b'\x01'              # nexthop_num
        b'\x01'              # nexthop_type
        b'\x00\x00\x00\x02'  # ifindex
    )
    family = socket.AF_INET6
    prefix = '2001:db8::/64'
    metric = 0x14
    nexthop_num = 1
    nexthop_type = zebra.ZEBRA_NEXTHOP_IFINDEX
    ifindex = 2

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraNexthopUpdate.parse(self.buf)

        self.assertEqual(self.family, body.family)
        self.assertEqual(self.prefix, body.prefix)
        self.assertEqual(self.metric, body.metric)
        self.assertEqual(self.nexthop_num, len(body.nexthops))
        self.assertEqual(self.nexthop_type, body.nexthops[0].type)
        self.assertEqual(self.ifindex, body.nexthops[0].ifindex)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraInterfaceNbrAddressAdd(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # ifindex
        b'\x02'              # family
        b'\xc0\xa8\x01\x00'  # prefix
        b'\x18'              # prefix_len
    )
    ifindex = 1
    family = socket.AF_INET
    prefix = '192.168.1.0/24'

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraInterfaceNbrAddressAdd.parse(self.buf)

        self.assertEqual(self.ifindex, body.ifindex)
        self.assertEqual(self.family, body.family)
        self.assertEqual(self.prefix, body.prefix)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraInterfaceBfdDestinationUpdate(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # ifindex
        b'\x02'              # dst_family
        b'\xc0\xa8\x01\x01'  # dst_prefix
        b'\x18'              # dst_prefix_len
        b'\x04'              # status
        b'\x02'              # src_family
        b'\xc0\xa8\x01\x02'  # src_prefix
        b'\x18'              # src_prefix_len
    )
    ifindex = 1
    dst_family = socket.AF_INET
    dst_prefix = '192.168.1.1/24'
    status = zebra.BFD_STATUS_UP
    src_family = socket.AF_INET
    src_prefix = '192.168.1.2/24'

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraInterfaceBfdDestinationUpdate.parse(self.buf)

        self.assertEqual(self.ifindex, body.ifindex)
        self.assertEqual(self.dst_family, body.dst_family)
        self.assertEqual(self.dst_prefix, body.dst_prefix)
        self.assertEqual(self.status, body.status)
        self.assertEqual(self.src_family, body.src_family)
        self.assertEqual(self.src_prefix, body.src_prefix)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraBfdDestinationRegisterMultiHopEnabled(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # pid
        b'\x00\x02'          # dst_family
        b'\xc0\xa8\x01\x01'  # dst_prefix
        b'\x00\x00\x00\x10'  # min_rx_timer
        b'\x00\x00\x00\x20'  # min_tx_timer
        b'\x01'              # detect_mult
        b'\x01'              # multi_hop
        b'\x00\x02'          # src_family
        b'\xc0\xa8\x01\x02'  # src_prefix
        b'\x05'              # multi_hop_count
    )
    pid = 1
    dst_family = socket.AF_INET
    dst_prefix = '192.168.1.1'
    min_rx_timer = 0x10
    min_tx_timer = 0x20
    detect_mult = 1
    multi_hop = 1
    src_family = socket.AF_INET
    src_prefix = '192.168.1.2'
    multi_hop_count = 5
    ifname = None

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraBfdDestinationRegister.parse(self.buf)

        self.assertEqual(self.pid, body.pid)
        self.assertEqual(self.dst_family, body.dst_family)
        self.assertEqual(self.dst_prefix, body.dst_prefix)
        self.assertEqual(self.min_rx_timer, body.min_rx_timer)
        self.assertEqual(self.min_tx_timer, body.min_tx_timer)
        self.assertEqual(self.detect_mult, body.detect_mult)
        self.assertEqual(self.multi_hop, body.multi_hop)
        self.assertEqual(self.src_family, body.src_family)
        self.assertEqual(self.src_prefix, body.src_prefix)
        self.assertEqual(self.multi_hop_count, body.multi_hop_count)
        self.assertEqual(self.ifname, body.ifname)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraBfdDestinationRegisterMultiHopDisabled(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # pid
        b'\x00\x02'          # dst_family
        b'\xc0\xa8\x01\x01'  # dst_prefix
        b'\x00\x00\x00\x10'  # min_rx_timer
        b'\x00\x00\x00\x20'  # min_tx_timer
        b'\x01'              # detect_mult
        b'\x00'              # multi_hop
        b'\x00\x02'          # src_family
        b'\xc0\xa8\x01\x02'  # src_prefix
        b'\x04'              # ifname_len
        b'eth0'              # ifname
    )
    pid = 1
    dst_family = socket.AF_INET
    dst_prefix = '192.168.1.1'
    min_rx_timer = 0x10
    min_tx_timer = 0x20
    detect_mult = 1
    multi_hop = 0
    src_family = socket.AF_INET
    src_prefix = '192.168.1.2'
    multi_hop_count = None
    ifname = 'eth0'

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraBfdDestinationRegister.parse(self.buf)

        self.assertEqual(self.pid, body.pid)
        self.assertEqual(self.dst_family, body.dst_family)
        self.assertEqual(self.dst_prefix, body.dst_prefix)
        self.assertEqual(self.min_rx_timer, body.min_rx_timer)
        self.assertEqual(self.min_tx_timer, body.min_tx_timer)
        self.assertEqual(self.detect_mult, body.detect_mult)
        self.assertEqual(self.multi_hop, body.multi_hop)
        self.assertEqual(self.src_family, body.src_family)
        self.assertEqual(self.src_prefix, body.src_prefix)
        self.assertEqual(self.multi_hop_count, body.multi_hop_count)
        self.assertEqual(self.ifname, body.ifname)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraBfdDestinationRegisterMultiHopEnabledIPv6(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # pid
        b'\x00\x0a'          # dst_family
        b'\x20\x01\x0d\xb8'  # dst_prefix
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x01'
        b'\x00\x00\x00\x10'  # min_rx_timer
        b'\x00\x00\x00\x20'  # min_tx_timer
        b'\x01'              # detect_mult
        b'\x01'              # multi_hop
        b'\x00\x0a'          # src_family
        b'\x20\x01\x0d\xb8'  # src_prefix
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x02'
        b'\x05'              # multi_hop_count
    )
    pid = 1
    dst_family = socket.AF_INET6
    dst_prefix = '2001:db8::1'
    min_rx_timer = 0x10
    min_tx_timer = 0x20
    detect_mult = 1
    multi_hop = 1
    src_family = socket.AF_INET6
    src_prefix = '2001:db8::2'
    multi_hop_count = 5
    ifname = None

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraBfdDestinationRegister.parse(self.buf)

        self.assertEqual(self.pid, body.pid)
        self.assertEqual(self.dst_family, body.dst_family)
        self.assertEqual(self.dst_prefix, body.dst_prefix)
        self.assertEqual(self.min_rx_timer, body.min_rx_timer)
        self.assertEqual(self.min_tx_timer, body.min_tx_timer)
        self.assertEqual(self.detect_mult, body.detect_mult)
        self.assertEqual(self.multi_hop, body.multi_hop)
        self.assertEqual(self.src_family, body.src_family)
        self.assertEqual(self.src_prefix, body.src_prefix)
        self.assertEqual(self.multi_hop_count, body.multi_hop_count)
        self.assertEqual(self.ifname, body.ifname)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraBfdDestinationDeregisterMultiHopEnabled(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # pid
        b'\x00\x02'          # dst_family
        b'\xc0\xa8\x01\x01'  # dst_prefix
        b'\x01'              # multi_hop
        b'\x00\x02'          # src_family
        b'\xc0\xa8\x01\x02'  # src_prefix
        b'\x05'              # multi_hop_count
    )
    pid = 1
    dst_family = socket.AF_INET
    dst_prefix = '192.168.1.1'
    multi_hop = 1
    src_family = socket.AF_INET
    src_prefix = '192.168.1.2'
    multi_hop_count = 5
    ifname = None

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraBfdDestinationDeregister.parse(self.buf)

        self.assertEqual(self.pid, body.pid)
        self.assertEqual(self.dst_family, body.dst_family)
        self.assertEqual(self.dst_prefix, body.dst_prefix)
        self.assertEqual(self.multi_hop, body.multi_hop)
        self.assertEqual(self.src_family, body.src_family)
        self.assertEqual(self.src_prefix, body.src_prefix)
        self.assertEqual(self.multi_hop_count, body.multi_hop_count)
        self.assertEqual(self.ifname, body.ifname)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraBfdDestinationDeregisterMultiHopDisabled(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # pid
        b'\x00\x02'          # dst_family
        b'\xc0\xa8\x01\x01'  # dst_prefix
        b'\x00'              # multi_hop
        b'\x00\x02'          # src_family
        b'\xc0\xa8\x01\x02'  # src_prefix
        b'\x04'              # ifname_len
        b'eth0'              # ifname
    )
    pid = 1
    dst_family = socket.AF_INET
    dst_prefix = '192.168.1.1'
    multi_hop = 0
    src_family = socket.AF_INET
    src_prefix = '192.168.1.2'
    multi_hop_count = None
    ifname = 'eth0'

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraBfdDestinationDeregister.parse(self.buf)

        self.assertEqual(self.pid, body.pid)
        self.assertEqual(self.dst_family, body.dst_family)
        self.assertEqual(self.dst_prefix, body.dst_prefix)
        self.assertEqual(self.multi_hop, body.multi_hop)
        self.assertEqual(self.src_family, body.src_family)
        self.assertEqual(self.src_prefix, body.src_prefix)
        self.assertEqual(self.multi_hop_count, body.multi_hop_count)
        self.assertEqual(self.ifname, body.ifname)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraBfdDestinationDeregisterMultiHopEnabledIPv6(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # pid
        b'\x00\x0a'          # dst_family
        b'\x20\x01\x0d\xb8'  # dst_prefix
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x01'
        b'\x01'              # multi_hop
        b'\x00\x0a'          # src_family
        b'\x20\x01\x0d\xb8'  # src_prefix
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x02'
        b'\x05'              # multi_hop_count
    )
    pid = 1
    dst_family = socket.AF_INET6
    dst_prefix = '2001:db8::1'
    multi_hop = 1
    src_family = socket.AF_INET6
    src_prefix = '2001:db8::2'
    multi_hop_count = 5
    ifname = None

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraBfdDestinationDeregister.parse(self.buf)

        self.assertEqual(self.pid, body.pid)
        self.assertEqual(self.dst_family, body.dst_family)
        self.assertEqual(self.dst_prefix, body.dst_prefix)
        self.assertEqual(self.multi_hop, body.multi_hop)
        self.assertEqual(self.src_family, body.src_family)
        self.assertEqual(self.src_prefix, body.src_prefix)
        self.assertEqual(self.multi_hop_count, body.multi_hop_count)
        self.assertEqual(self.ifname, body.ifname)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraVrfAdd(unittest.TestCase):
    buf = (
        b'VRF1'              # vrf_name
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
    )
    vrf_name = 'VRF1'

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraVrfAdd.parse(self.buf)

        self.assertEqual(self.vrf_name, body.vrf_name)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraInterfaceVrfUpdate(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # ifindex
        b'\x00\x02'          # vrf_id
    )
    ifindex = 1
    vrf_id = 2

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraInterfaceVrfUpdate.parse(self.buf)

        self.assertEqual(self.ifindex, body.ifindex)
        self.assertEqual(self.vrf_id, body.vrf_id)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraInterfaceEnableRadv(unittest.TestCase):
    buf = (
        b'\x00\x00\x00\x01'  # ifindex
        b'\x00\x00\x01\x00'  # interval
    )
    ifindex = 1
    interval = 0x100

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraInterfaceEnableRadv.parse(self.buf)

        self.assertEqual(self.ifindex, body.ifindex)
        self.assertEqual(self.interval, body.interval)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraMplsLabelsAddIPv4(unittest.TestCase):
    buf = (
        b'\x09'              # route_type
        b'\x00\x00\x00\x02'  # family
        b'\xc0\xa8\x01\x00'  # prefix
        b'\x18'              # prefix_len
        b'\xc0\xa8\x01\x01'  # gate_addr
        b'\x10'              # distance
        b'\x00\x00\x00\x64'  # in_label
        b'\x00\x00\x00\x03'  # out_label
    )
    route_type = zebra.ZEBRA_ROUTE_BGP
    family = socket.AF_INET
    prefix = '192.168.1.0/24'
    gate_addr = '192.168.1.1'
    distance = 0x10
    in_label = 100
    out_label = zebra.MPLS_IMP_NULL_LABEL

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraMplsLabelsAdd.parse(self.buf)

        self.assertEqual(self.route_type, body.route_type)
        self.assertEqual(self.family, body.family)
        self.assertEqual(self.prefix, body.prefix)
        self.assertEqual(self.gate_addr, body.gate_addr)
        self.assertEqual(self.distance, body.distance)
        self.assertEqual(self.in_label, body.in_label)
        self.assertEqual(self.out_label, body.out_label)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))


class TestZebraMplsLabelsAddIPv6(unittest.TestCase):
    buf = (
        b'\x09'              # route_type
        b'\x00\x00\x00\x0a'  # family
        b'\x20\x01\x0d\xb8'  # prefix
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x40'              # prefix_len
        b'\x20\x01\x0d\xb8'  # gate_addr
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x00'
        b'\x00\x00\x00\x01'
        b'\x10'              # distance
        b'\x00\x00\x00\x64'  # in_label
        b'\x00\x00\x00\x03'  # out_label
    )
    route_type = zebra.ZEBRA_ROUTE_BGP
    family = socket.AF_INET6
    prefix = '2001:db8::/64'
    gate_addr = '2001:db8::1'
    distance = 0x10
    in_label = 100
    out_label = zebra.MPLS_IMP_NULL_LABEL

    @_patch_frr_v2
    def test_parser(self):
        body = zebra.ZebraMplsLabelsAdd.parse(self.buf)

        self.assertEqual(self.route_type, body.route_type)
        self.assertEqual(self.family, body.family)
        self.assertEqual(self.prefix, body.prefix)
        self.assertEqual(self.gate_addr, body.gate_addr)
        self.assertEqual(self.distance, body.distance)
        self.assertEqual(self.in_label, body.in_label)
        self.assertEqual(self.out_label, body.out_label)

        buf = body.serialize()

        self.assertEqual(binary_str(self.buf), binary_str(buf))



# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Zebra Server corresponding to 'zserv' structure.
"""

import contextlib
import logging
import os
import socket
import struct

from os_ken import cfg
from os_ken.base import app_manager
from os_ken.base.app_manager import OSKenApp
from os_ken.controller.handler import set_ev_cls
from os_ken.lib import hub
from os_ken.lib import ip
from os_ken.lib.packet import zebra

from os_ken.services.protocols.zebra import db
from os_ken.services.protocols.zebra import event
from os_ken.services.protocols.zebra.server import event as zserver_event


LOG = logging.getLogger(__name__)

CONF = cfg.CONF['zapi']
GLOBAL_CONF = cfg.CONF

# Session to database of Zebra protocol service
SESSION = db.Session()


class ZClient(object):
    """
    Zebra client class.
    """

    def __init__(self, server, sock, addr):
        self.server = server
        self.sock = sock
        self.addr = addr
        self.logger = server.logger
        self.is_active = False
        self._threads = []
        self.send_q = hub.Queue(16)

        # Zebra protocol version
        self.zserv_ver = CONF.server_version

        # Zebra route type distributed by client (not initialized yet)
        self.route_type = None

    def start(self):
        self.is_active = True
        self.sock.settimeout(GLOBAL_CONF.socket_timeout)

        self._threads.append(hub.spawn(self._send_loop))
        self._threads.append(hub.spawn(self._recv_loop))

        self.server.send_event_to_observers(
            zserver_event.EventZClientConnected(self))

        hub.joinall(self._threads)

        self.server.send_event_to_observers(
            zserver_event.EventZClientDisconnected(self))

    def stop(self):
        self.is_active = False

    def _send_loop(self):
        try:
            while self.is_active:
                buf = self.send_q.get()
                self.sock.sendall(buf)
        except socket.error as e:
            self.logger.exception(
                'Error while sending message to Zebra client%s: %s',
                self.addr, e)

        self.stop()

    def _recv_loop(self):
        buf = b''
        min_len = recv_len = zebra.ZebraMessage.get_header_size(
            self.zserv_ver)
        try:
            while self.is_active:
                try:
                    recv_buf = self.sock.recv(recv_len)
                except socket.timeout:
                    continue

                if len(recv_buf) == 0:
                    break

                buf += recv_buf
                while len(buf) >= min_len:
                    (length,) = struct.unpack_from('!H', buf)
                    if (length - len(buf)) > 0:
                        # Need to receive remaining data
                        recv_len = length - len(buf)
                        break

                    msg, _, buf = zebra.ZebraMessage.parser(buf)

                    ev = event.message_to_event(self, msg)
                    if ev:
                        self.logger.debug('Notify event: %s', ev)
                        self.server.send_event_to_observers(ev)

        except socket.error as e:
            self.logger.exception(
                'Error while sending message to Zebra client%s: %s',
                self.addr, e)

        self.stop()

    def send_msg(self, msg):
        """
        Sends Zebra message.

        :param msg: Instance of py:class: `os_ken.lib.packet.zebra.ZebraMessage`.
        :return: Serialized msg if succeeded, otherwise None.
        """
        if not self.is_active:
            self.logger.debug(
                'Cannot send message: Already deactivated: msg=%s', msg)
            return
        elif not self.send_q:
            self.logger.debug(
                'Cannot send message: Send queue does not exist: msg=%s', msg)
            return
        elif self.zserv_ver != msg.version:
            self.logger.debug(
                'Zebra protocol version mismatch:'
                'server_version=%d, msg.version=%d',
                self.zserv_ver, msg.version)
            msg.version = self.zserv_ver  # fixup

        self.send_q.put(msg.serialize())


def zclient_connection_factory(sock, addr):
    LOG.debug('Connected from client: %s: %s', addr, sock)
    zserv = app_manager.lookup_service_brick(ZServer.__name__)
    with contextlib.closing(ZClient(zserv, sock, addr)) as zclient:
        try:
            zclient.start()
        except Exception as e:
            LOG.error('Error in client%s: %s', addr, e)
            raise e


def detect_address_family(host):
    if ip.valid_ipv4(host):
        return socket.AF_INET
    elif ip.valid_ipv6(host):
        return socket.AF_INET6
    elif os.path.isdir(os.path.dirname(host)):
        return socket.AF_UNIX
    else:
        return None


class ZServer(OSKenApp):
    """
    The base class for Zebra server application.
    """
    _EVENTS = event.ZEBRA_EVENTS + [
        zserver_event.EventZClientConnected,
        zserver_event.EventZClientDisconnected,
    ]

    def __init__(self, *args, **kwargs):
        super(ZServer, self).__init__(*args, **kwargs)
        self.zserv = None
        self.zserv_addr = (CONF.server_host, CONF.server_port)
        self.zapi_connection_family = detect_address_family(CONF.server_host)

        # Initial Router ID for Zebra server
        self.router_id = CONF.router_id

    def start(self):
        super(ZServer, self).start()

        if self.zapi_connection_family == socket.AF_UNIX:
            unix_sock_dir = os.path.dirname(CONF.server_host)
            # Makes sure the unix socket does not already exist
            if os.path.exists(CONF.server_host):
                os.remove(CONF.server_host)
            if not os.path.isdir(unix_sock_dir):
                os.mkdir(unix_sock_dir)
                os.chmod(unix_sock_dir, 0o777)

        try:
            self.zserv = hub.StreamServer(
                self.zserv_addr, zclient_connection_factory)
        except OSError as e:
            self.logger.error(
                'Cannot start Zebra server%s: %s', self.zserv_addr, e)
            raise e

        if self.zapi_connection_family == socket.AF_UNIX:
            os.chmod(CONF.server_host, 0o777)

        self._add_lo_interface()

        return hub.spawn(self.zserv.serve_forever)

    def _add_lo_interface(self):
        intf = db.interface.ip_link_add(SESSION, 'lo')
        if intf:
            self.logger.debug('Added interface "%s": %s', intf.ifname, intf)

        route = db.route.ip_route_add(
            SESSION,
            destination='127.0.0.0/8',
            device='lo',
            source='127.0.0.1/8',
            route_type=zebra.ZEBRA_ROUTE_CONNECT)
        if route:
            self.logger.debug(
                'Added route to "%s": %s', route.destination, route)

    @set_ev_cls(event.EventZebraHello)
    def _hello_handler(self, ev):
        if ev.body is None:
            self.logger.debug('Client %s says hello.', ev.zclient)
            return

        # Set distributed route_type to ZClient
        ev.zclient.route_type = ev.body.route_type
        self.logger.debug(
            'Client %s says hello and bids fair to announce only %s routes',
            ev.zclient, ev.body.route_type)

    @set_ev_cls(event.EventZebraRouterIDAdd)
    def _router_id_add_handler(self, ev):
        self.logger.debug(
            'Client %s requests router_id, server will response: router_id=%s',
            ev.zclient, self.router_id)

        # Send ZEBRA_ROUTER_ID_UPDATE for response
        msg = zebra.ZebraMessage(
            body=zebra.ZebraRouterIDUpdate(
                family=socket.AF_INET,
                prefix='%s/32' % self.router_id))
        ev.zclient.send_msg(msg)

    @set_ev_cls(event.EventZebraInterfaceAdd)
    def _interface_add_handler(self, ev):
        self.logger.debug('Client %s requested all interfaces', ev.zclient)

        interfaces = db.interface.ip_address_show_all(SESSION)
        self.logger.debug('Server will response interfaces: %s', interfaces)
        for intf in interfaces:
            msg = zebra.ZebraMessage(
                body=zebra.ZebraInterfaceAdd(
                    ifname=intf.ifname,
                    ifindex=intf.ifindex,
                    status=intf.status,
                    if_flags=intf.flags,
                    ptm_enable=zebra.ZEBRA_IF_PTM_ENABLE_OFF,
                    ptm_status=zebra.ZEBRA_PTM_STATUS_UNKNOWN,
                    metric=intf.metric,
                    ifmtu=intf.ifmtu,
                    ifmtu6=intf.ifmtu6,
                    bandwidth=intf.bandwidth,
                    ll_type=intf.ll_type,
                    hw_addr=intf.hw_addr))
            ev.zclient.send_msg(msg)

            routes = db.route.ip_route_show_all(
                SESSION, ifindex=intf.ifindex, is_selected=True)
            self.logger.debug('Server will response routes: %s', routes)
            for route in routes:
                dest, _ = route.destination.split('/')
                msg = zebra.ZebraMessage(
                    body=zebra.ZebraInterfaceAddressAdd(
                        ifindex=intf.ifindex,
                        ifc_flags=0,
                        family=None,
                        prefix=route.source,
                        dest=dest))
                ev.zclient.send_msg(msg)

    @set_ev_cls([event.EventZebraIPv4RouteAdd,
                 event.EventZebraIPv6RouteAdd])
    def _ip_route_add_handler(self, ev):
        self.logger.debug(
            'Client %s advertised IP route: %s', ev.zclient, ev.body)

        for nexthop in ev.body.nexthops:
            route = db.route.ip_route_add(
                SESSION,
                destination=ev.body.prefix,
                gateway=nexthop.addr,
                ifindex=nexthop.ifindex or 0,
                route_type=ev.body.route_type)
            if route:
                self.logger.debug(
                    'Added route to "%s": %s', route.destination, route)

    @set_ev_cls([event.EventZebraIPv4RouteDelete,
                 event.EventZebraIPv6RouteDelete])
    def _ip_route_delete_handler(self, ev):
        self.logger.debug(
            'Client %s withdrew IP route: %s', ev.zclient, ev.body)

        for nexthop in ev.body.nexthops:
            routes = db.route.ip_route_delete(
                SESSION,
                destination=ev.body.prefix,
                gateway=nexthop.addr,
                route_type=ev.body.route_type)
            if routes:
                self.logger.debug(
                    'Deleted routes to "%s": %s', ev.body.prefix, routes)



# Copyright (C) 2015 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 YAMAMOTO Takashi <yamamoto at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import os_ken.ofproto.ofproto_v1_3_parser as ofpp


class Test_FlowSpec(unittest.TestCase):
    def test_flowspec_src_0_dst_0(self):
        user = ofpp.NXFlowSpecMatch(src=('in_port', 0),
                                    dst=('in_port', 0),
                                    n_bits=16)
        on_wire = (
            b'\x00\x10'
            b'\x80\x00\x00\x04\x00\x00'
            b'\x80\x00\x00\x04\x00\x00'
        )
        self.assertEqual(on_wire, user.serialize())
        (o, rest) = ofpp._NXFlowSpec.parse(on_wire)
        self.assertEqual(user.to_jsondict(), o.to_jsondict())
        self.assertEqual(str(user), str(o))
        self.assertEqual(b'', rest)

    def test_flowspec_src_1_dst_0(self):
        user = ofpp.NXFlowSpecMatch(src=99,
                                    dst=('in_port', 0),
                                    n_bits=16)
        on_wire = (
            b'\x20\x10'
            b'\x00\x63'
            b'\x80\x00\x00\x04\x00\x00'
        )
        self.assertEqual(on_wire, user.serialize())
        (o, rest) = ofpp._NXFlowSpec.parse(on_wire)
        self.assertEqual(user.to_jsondict(), o.to_jsondict())
        self.assertEqual(str(user), str(o))
        self.assertEqual(b'', rest)

    def test_flowspec_src_0_dst_1(self):
        user = ofpp.NXFlowSpecLoad(src=('in_port', 0),
                                   dst=('in_port', 0),
                                   n_bits=16)
        on_wire = (
            b'\x08\x10'
            b'\x80\x00\x00\x04\x00\x00'
            b'\x80\x00\x00\x04\x00\x00'
        )
        self.assertEqual(on_wire, user.serialize())
        (o, rest) = ofpp._NXFlowSpec.parse(on_wire)
        self.assertEqual(user.to_jsondict(), o.to_jsondict())
        self.assertEqual(str(user), str(o))
        self.assertEqual(b'', rest)

    def test_flowspec_src_1_dst_1(self):
        user = ofpp.NXFlowSpecLoad(src=99,
                                   dst=('in_port', 0),
                                   n_bits=16)
        on_wire = (
            b'\x28\x10'
            b'\x00\x63'
            b'\x80\x00\x00\x04\x00\x00'
        )
        self.assertEqual(on_wire, user.serialize())
        (o, rest) = ofpp._NXFlowSpec.parse(on_wire)
        self.assertEqual(user.to_jsondict(), o.to_jsondict())
        self.assertEqual(str(user), str(o))
        self.assertEqual(b'', rest)

    def test_flowspec_src_0_dst_2(self):
        user = ofpp.NXFlowSpecOutput(src=('in_port', 0),
                                     dst='',
                                     n_bits=16)
        on_wire = (
            b'\x10\x10'
            b'\x80\x00\x00\x04\x00\x00'
        )
        self.assertEqual(on_wire, user.serialize())
        (o, rest) = ofpp._NXFlowSpec.parse(on_wire)
        self.assertEqual(user.to_jsondict(), o.to_jsondict())
        self.assertEqual(str(user), str(o))
        self.assertEqual(b'', rest)



# Copyright (C) 2012 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# vim: tabstop=4 shiftwidth=4 softtabstop=4

import unittest
import logging
from os_ken.ofproto.ether import *


LOG = logging.getLogger('test_ether')


class TestInet(unittest.TestCase):
    """ Test case for ether
    """

    def test_ether_type(self):
        self.assertEqual(ETH_TYPE_IP, 0x0800)
        self.assertEqual(ETH_TYPE_ARP, 0x0806)
        self.assertEqual(ETH_TYPE_8021Q, 0x8100)
        self.assertEqual(ETH_TYPE_IPV6, 0x86dd)
        self.assertEqual(ETH_TYPE_MPLS, 0x8847)
        self.assertEqual(ETH_TYPE_SLOW, 0x8809)



# Copyright (C) 2017 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import os_ken.exception
from os_ken.lib.ofctl_utils import str_to_int
from os_ken.ofproto import nicira_ext


def ofp_instruction_from_str(ofproto, action_str):
    """
    Parse an ovs-ofctl style action string and return a list of
    jsondict representations of OFPInstructionActions, which
    can then be passed to ofproto_parser.ofp_instruction_from_jsondict.

    Please note that this is for making transition from ovs-ofctl
    easier. Please consider using OFPAction constructors when writing
    new codes.

    This function takes the following arguments.

    =========== =================================================
    Argument    Description
    =========== =================================================
    ofproto     An ofproto module.
    action_str  An action string.
    =========== =================================================
    """
    action_re = re.compile(r"([a-z_]+)(\([^)]*\)|[^a-z_,()][^,()]*)*")
    result = []
    while len(action_str):
        m = action_re.match(action_str)
        if not m:
            raise os_ken.exception.OFPInvalidActionString(action_str=action_str)
        action_name = m.group(1)
        this_action = m.group(0)
        paren_level = this_action.count('(') - this_action.count(')')
        assert paren_level >= 0
        try:
            # Parens can be nested. Look for as many ')'s as '('s.
            if paren_level > 0:
                this_action, rest = _tokenize_paren_block(action_str, m.end(0))
            else:
                rest = action_str[m.end(0):]
            if len(rest):
                assert rest[0] == ','
                rest = rest[1:]
        except Exception:
            raise os_ken.exception.OFPInvalidActionString(action_str=action_str)
        if action_name == 'drop':
            assert this_action == 'drop'
            assert len(result) == 0 and rest == ''
            return []
        converter = getattr(OfctlActionConverter, action_name, None)
        if converter is None or not callable(converter):
            raise os_ken.exception.OFPInvalidActionString(action_str=action_name)
        result.append(converter(ofproto, this_action))
        action_str = rest

    return result


def _tokenize_paren_block(string, pos):
    paren_re = re.compile("[()]")
    paren_level = string[:pos].count('(') - string[:pos].count(')')
    while paren_level > 0:
        m = paren_re.search(string[pos:])
        if m.group(0) == '(':
            paren_level += 1
        else:
            paren_level -= 1
        pos += m.end(0)
    return string[:pos], string[pos:]


def tokenize_ofp_instruction_arg(arg):
    """
    Tokenize an argument portion of ovs-ofctl style action string.
    """
    arg_re = re.compile("[^,()]*")
    try:
        rest = arg
        result = []
        while len(rest):
            m = arg_re.match(rest)
            if m.end(0) == len(rest):
                result.append(rest)
                return result
            if rest[m.end(0)] == '(':
                this_block, rest = _tokenize_paren_block(
                    rest, m.end(0) + 1)
                result.append(this_block)
            elif rest[m.end(0)] == ',':
                result.append(m.group(0))
                rest = rest[m.end(0):]
            else:  # is ')'
                raise Exception
            if len(rest):
                assert rest[0] == ','
                rest = rest[1:]
        return result
    except Exception:
        raise os_ken.exception.OFPInvalidActionString(action_str=arg)


_OXM_FIELD_OFCTL_ALIASES = {
    'tun_id': 'tunnel_id',
    'in_port': 'in_port_nxm',
    'in_port_oxm': 'in_port',
    'dl_src': 'eth_src',
    'dl_type': 'eth_type',
    'nw_src': 'ipv4_src',
    'ip_src': 'ipv4_src',
    'nw_proto': 'ip_proto',
    'nw_ecn': 'ip_ecn',
    'tp_src': 'tcp_src',
    'icmp_type': 'icmpv4_type',
    'icmp_code': 'icmpv4_code',
    'nd_target': 'ipv6_nd_target',
    'nd_sll': 'ipv6_nd_sll',
    'nd_tll': 'ipv6_nd_tll',
    # Nicira extension
    'tun_src': 'tun_ipv4_src'
}


def ofp_ofctl_field_name_to_os_ken(field):
    """Convert an ovs-ofctl field name to os_ken equivalent."""
    mapped = _OXM_FIELD_OFCTL_ALIASES.get(field)
    if mapped:
        return mapped
    if field.endswith("_dst"):
        mapped = _OXM_FIELD_OFCTL_ALIASES.get(field[:-3] + "src")
        if mapped:
            return mapped[:-3] + "dst"
    return field


_NXM_FIELD_MAP = dict([(key, key + '_nxm') for key in [
    'arp_sha', 'arp_tha', 'ipv6_src', 'ipv6_dst',
    'icmpv6_type', 'icmpv6_code', 'ip_ecn', 'tcp_flags']])
_NXM_FIELD_MAP.update({
    'tun_id': 'tunnel_id_nxm', 'ip_ttl': 'nw_ttl'})

_NXM_OF_FIELD_MAP = dict([(key, key + '_nxm') for key in [
    'in_port', 'eth_dst', 'eth_src', 'eth_type', 'ip_proto',
    'tcp_src', 'tcp_dst', 'udp_src', 'udp_dst',
    'arp_op', 'arp_spa', 'arp_tpa']])
_NXM_OF_FIELD_MAP.update({
    'ip_src': 'ipv4_src_nxm', 'ip_dst': 'ipv4_dst_nxm',
    'icmp_type': 'icmpv4_type_nxm', 'icmp_code': 'icmpv4_code_nxm'})


def nxm_field_name_to_os_ken(field):
    """
    Convert an ovs-ofctl style NXM_/OXM_ field name to
    a os_ken match field name.
    """
    if field.endswith("_W"):
        field = field[:-2]
    prefix = field[:7]
    field = field[7:].lower()
    mapped_result = None

    if prefix == 'NXM_NX_':
        mapped_result = _NXM_FIELD_MAP.get(field)
    elif prefix == "NXM_OF_":
        mapped_result = _NXM_OF_FIELD_MAP.get(field)
    elif prefix == "OXM_OF_":
        # no mapping needed
        pass
    else:
        raise ValueError
    if mapped_result is not None:
        return mapped_result
    return field


class OfctlActionConverter(object):

    @classmethod
    def goto_table(cls, ofproto, action_str):
        assert action_str.startswith('goto_table:')
        table_id = str_to_int(action_str[len('goto_table:'):])
        return dict(OFPInstructionGotoTable={'table_id': table_id})

    @classmethod
    def normal(cls, ofproto, action_str):
        return cls.output(ofproto, action_str)

    @classmethod
    def output(cls, ofproto, action_str):
        if action_str == 'normal':
            port = ofproto.OFPP_NORMAL
        else:
            assert action_str.startswith('output:')
            port = str_to_int(action_str[len('output:'):])
        return dict(OFPActionOutput={'port': port})

    @classmethod
    def pop_vlan(cls, ofproto, action_str):
        return dict(OFPActionPopVlan={})

    @classmethod
    def set_field(cls, ofproto, action_str):
        try:
            assert action_str.startswith("set_field:")
            value, key = action_str[len("set_field:"):].split("->", 1)
            fieldarg = dict(field=ofp_ofctl_field_name_to_os_ken(key))
            m = value.find('/')
            if m >= 0:
                fieldarg['value'] = str_to_int(value[:m])
                fieldarg['mask'] = str_to_int(value[m + 1:])
            else:
                fieldarg['value'] = str_to_int(value)
        except Exception:
            raise os_ken.exception.OFPInvalidActionString(action_str=action_str)
        return dict(OFPActionSetField={
            'field': {'OXMTlv': fieldarg}})

    # NX actions
    @classmethod
    def resubmit(cls, ofproto, action_str):
        arg = action_str[len("resubmit"):]
        kwargs = {}
        try:
            if arg[0] == ':':
                kwargs['in_port'] = str_to_int(arg[1:])
            elif arg[0] == '(' and arg[-1] == ')':
                in_port, table_id = arg[1:-1].split(',')
                if in_port:
                    kwargs['in_port'] = str_to_int(in_port)
                if table_id:
                    kwargs['table_id'] = str_to_int(table_id)
            else:
                raise Exception
            return dict(NXActionResubmitTable=kwargs)
        except Exception:
            raise os_ken.exception.OFPInvalidActionString(
                action_str=action_str)

    @classmethod
    def conjunction(cls, ofproto, action_str):
        try:
            assert action_str.startswith('conjunction(')
            assert action_str[-1] == ')'
            args = action_str[len('conjunction('):-1].split(',')
            assert len(args) == 2
            id_ = str_to_int(args[0])
            clauses = list(map(str_to_int, args[1].split('/')))
            assert len(clauses) == 2
            return dict(NXActionConjunction={
                'clause': clauses[0] - 1,
                'n_clauses': clauses[1],
                'id': id_})
        except Exception:
            raise os_ken.exception.OFPInvalidActionString(
                action_str=action_str)

    @classmethod
    def ct(cls, ofproto, action_str):
        str_to_port = {'ftp': 21, 'tftp': 69}
        flags = 0
        zone_src = ""
        zone_ofs_nbits = 0
        recirc_table = nicira_ext.NX_CT_RECIRC_NONE
        alg = 0
        ct_actions = []

        if len(action_str) > 2:
            if (not action_str.startswith('ct(') or
                    action_str[-1] != ')'):
                raise os_ken.exception.OFPInvalidActionString(
                    action_str=action_str)
            rest = tokenize_ofp_instruction_arg(action_str[len('ct('):-1])
        else:
            rest = []
        for arg in rest:
            if arg == 'commit':
                flags |= nicira_ext.NX_CT_F_COMMIT
                rest = rest[len('commit'):]
            elif arg == 'force':
                flags |= nicira_ext.NX_CT_F_FORCE
            elif arg.startswith('exec('):
                ct_actions = ofp_instruction_from_str(
                    ofproto, arg[len('exec('):-1])
            else:
                try:
                    k, v = arg.split('=', 1)
                    if k == 'table':
                        recirc_table = str_to_int(v)
                    elif k == 'zone':
                        m = re.search(r'\[(\d*)\.\.(\d*)\]', v)
                        if m:
                            zone_ofs_nbits = nicira_ext.ofs_nbits(
                                int(m.group(1)), int(m.group(2)))
                            zone_src = nxm_field_name_to_os_ken(
                                v[:m.start(0)])
                        else:
                            zone_ofs_nbits = str_to_int(v)
                    elif k == 'alg':
                        alg = str_to_port[arg[len('alg='):]]
                except Exception:
                    raise os_ken.exception.OFPInvalidActionString(
                        action_str=action_str)
        return dict(NXActionCT={'flags': flags,
                                'zone_src': zone_src,
                                'zone_ofs_nbits': zone_ofs_nbits,
                                'recirc_table': recirc_table,
                                'alg': alg,
                                'actions': ct_actions})

    @classmethod
    def ct_clear(cls, ofproto, action_str):
        return dict(NXActionCTClear={})



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 Module provides utilities for validation.
"""
import numbers
import re
import socket

from os_ken.lib import ip


def is_valid_mac(mac):
    """Returns True if the given MAC address is valid.

    The given MAC address should be a colon hexadecimal notation string.

    Samples:
        - valid address: aa:bb:cc:dd:ee:ff, 11:22:33:44:55:66
        - invalid address: aa:bb:cc:dd, 11-22-33-44-55-66, etc.
    """
    return bool(re.match(r'^' + r'[\:\-]'.join([r'([0-9a-f]{2})'] * 6)
                         + r'$', mac.lower()))


def is_valid_ip_prefix(prefix, bits):
    """Returns True if *prefix* is a valid IPv4 or IPv6 address prefix.

    *prefix* should be a number between 0 to *bits* length.
    """
    try:
        # Prefix should be a number
        prefix = int(prefix)
    except ValueError:
        return False

    # Prefix should be a number between 0 to *bits*
    return 0 <= prefix <= bits


def is_valid_ipv4(ipv4):
    """Returns True if given is a valid ipv4 address.

    Given value should be a dot-decimal notation string.

    Samples:
        - valid address: 10.0.0.1, 192.168.0.1
        - invalid address: 11.0.0, 192:168:0:1, etc.
    """
    return ip.valid_ipv4(ipv4)


def is_valid_ipv4_prefix(ipv4_prefix):
    """Returns True if *ipv4_prefix* is a valid prefix with mask.

    Samples:
        - valid prefix: 1.1.1.0/32, 244.244.244.1/10
        - invalid prefix: 255.2.2.2/2, 2.2.2/22, etc.
    """
    if not isinstance(ipv4_prefix, str):
        return False

    tokens = ipv4_prefix.split('/')
    if len(tokens) != 2:
        return False

    # Validate address/mask and return
    return is_valid_ipv4(tokens[0]) and is_valid_ip_prefix(tokens[1], 32)


def is_valid_ipv6(ipv6):
    """Returns True if given `ipv6` is a valid IPv6 address
    """
    return ip.valid_ipv6(ipv6)


def is_valid_ipv6_prefix(ipv6_prefix):
    """Returns True if given `ipv6_prefix` is a valid IPv6 prefix."""

    # Validate input type
    if not isinstance(ipv6_prefix, str):
        return False

    tokens = ipv6_prefix.split('/')
    if len(tokens) != 2:
        return False

    # Validate address/mask and return
    return is_valid_ipv6(tokens[0]) and is_valid_ip_prefix(tokens[1], 128)


def is_valid_old_asn(asn):
    """Returns True if the given AS number is Two Octet."""
    return isinstance(asn, numbers.Integral) and 0 <= asn <= 0xffff


def is_valid_asn(asn):
    """Returns True if the given AS number is Two or Four Octet."""
    return isinstance(asn, numbers.Integral) and 0 <= asn <= 0xffffffff


def is_valid_vpnv4_prefix(prefix):
    """Returns True if given prefix is a string represent vpnv4 prefix.

    Vpnv4 prefix is made up of RD:Ipv4, where RD is represents route
    distinguisher and Ipv4 represents valid dot-decimal ipv4 notation string.
    """
    if not isinstance(prefix, str):
        return False

    # Split the prefix into route distinguisher and IP
    tokens = prefix.split(':', 2)
    if len(tokens) != 3:
        return False

    # Validate route distinguisher
    if not is_valid_route_dist(':'.join([tokens[0], tokens[1]])):
        return False

    # Validate IPv4 prefix and return
    return is_valid_ipv4_prefix(tokens[2])


def is_valid_vpnv6_prefix(prefix):
    """Returns True if given prefix is a string represent vpnv6 prefix.

    Vpnv6 prefix is made up of RD:Ipv6, where RD is represents route
    distinguisher and Ipv6 represents valid colon hexadecimal notation string.
    """
    if not isinstance(prefix, str):
        return False

    # Split the prefix into route distinguisher and IP
    tokens = prefix.split(':', 2)
    if len(tokens) != 3:
        return False

    # Validate route distinguisher
    if not is_valid_route_dist(':'.join([tokens[0], tokens[1]])):
        return False

    # Validate IPv6 prefix and return
    return is_valid_ipv6_prefix(tokens[2])


def is_valid_med(med):
    """Returns True if value of *med* is valid as per RFC.

    According to RFC MED is a four octet non-negative integer and
    value '((2 ** 32) - 1) =  0xffffffff' denotes an "infinity" metric.
    """
    return isinstance(med, numbers.Integral) and 0 <= med <= 0xffffffff


def is_valid_mpls_label(label):
    """Validates `label` according to MPLS label rules

    RFC says:
    This 20-bit field.
    A value of 0 represents the "IPv4 Explicit NULL Label".
    A value of 1 represents the "Router Alert Label".
    A value of 2 represents the "IPv6 Explicit NULL Label".
    A value of 3 represents the "Implicit NULL Label".
    Values 4-15 are reserved.
    """
    if (not isinstance(label, numbers.Integral) or
            (4 <= label <= 15) or
            (label < 0 or label > 2 ** 20)):
        return False

    return True


def is_valid_mpls_labels(labels):
    """Returns True if the given value is a list of valid MPLS labels.
    """
    if not isinstance(labels, (list, tuple)):
        return False

    for label in labels:
        if not is_valid_mpls_label(label):
            return False

    return True


def is_valid_route_dist(route_dist):
    """Validates *route_dist* as string representation of route distinguisher.

    Returns True if *route_dist* is as per our convention of RD, else False.
    Our convention is to represent RD as a string in format:
    *admin_sub_field:assigned_num_field* and *admin_sub_field* can be valid
    IPv4 string representation.
    Valid examples: '65000:222', '1.2.3.4:4432'.
    Invalid examples: '1.11.1: 333'
    """
    # TODO(PH): Provide complete implementation.
    return is_valid_ext_comm_attr(route_dist)


def is_valid_ext_comm_attr(attr):
    """Validates *attr* as string representation of RT or SOO.

    Returns True if *attr* is as per our convention of RT or SOO, else
    False. Our convention is to represent RT/SOO is a string with format:
    *global_admin_part:local_admin_path*
    """
    if not isinstance(attr, str):
        return False

    tokens = attr.rsplit(':', 1)
    if len(tokens) != 2:
        return False

    try:
        if '.' in tokens[0]:
            if not is_valid_ipv4(tokens[0]):
                return False
        else:
            int(tokens[0])
        int(tokens[1])
    except (ValueError, socket.error):
        return False

    return True


def is_valid_esi(esi):
    """Returns True if the given EVPN Ethernet SegmentEthernet ID is valid."""
    if isinstance(esi, numbers.Integral):
        return 0 <= esi <= 0xffffffffffffffffff
    return isinstance(esi, dict)


def is_valid_ethernet_tag_id(etag_id):
    """Returns True if the given EVPN Ethernet Tag ID is valid.

    Ethernet Tag ID should be a 32-bit field number.
    """
    return isinstance(etag_id, numbers.Integral) and 0 <= etag_id <= 0xffffffff


def is_valid_vni(vni):
    """Returns True if the given Virtual Network Identifier for VXLAN
    is valid.

    Virtual Network Identifier should be a 24-bit field number.
    """
    return isinstance(vni, numbers.Integral) and 0 <= vni <= 0xffffff



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 Isaku Yamahata <yamahata at private email ne jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import logging
import struct
import inspect

from os_ken.ofproto import inet
from os_ken.lib.packet import ipv4
from os_ken.lib.packet import ipv6
from os_ken.lib.packet import packet
from os_ken.lib.packet import packet_utils
from os_ken.lib.packet import vrrp
from os_ken.lib import addrconv


LOG = logging.getLogger(__name__)


class Test_vrrpv2(unittest.TestCase):
    """ Test case for vrrp v2
    """
    version = vrrp.VRRP_VERSION_V2
    type_ = vrrp.VRRP_TYPE_ADVERTISEMENT
    vrid = 128
    priority = 100
    count_ip = 1
    auth_type = vrrp.VRRP_AUTH_NO_AUTH
    max_adver_int = 100
    checksum = 0
    ip_address = '192.168.0.1'
    auth_data = (0, 0)
    vrrpv2 = vrrp.vrrpv2.create(type_, vrid, priority, max_adver_int,
                                [ip_address])
    buf = struct.pack(vrrp.vrrpv2._PACK_STR + '4sII',
                      vrrp.vrrp_to_version_type(vrrp.VRRP_VERSION_V2, type_),
                      vrid, priority, count_ip,
                      auth_type, max_adver_int, checksum,
                      addrconv.ipv4.text_to_bin(ip_address),
                      auth_data[0], auth_data[1])

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.type_, self.vrrpv2.type)
        self.assertEqual(self.vrid, self.vrrpv2.vrid)
        self.assertEqual(self.priority, self.vrrpv2.priority)
        self.assertEqual(self.count_ip, self.vrrpv2.count_ip)
        self.assertEqual(self.auth_type, self.vrrpv2.auth_type)
        self.assertEqual(1, len(self.vrrpv2.ip_addresses))
        self.assertEqual(self.ip_address, self.vrrpv2.ip_addresses[0])
        self.assertEqual(self.auth_data, self.vrrpv2.auth_data)

    def test_parser(self):
        vrrpv2, _cls, _ = self.vrrpv2.parser(self.buf)

        self.assertEqual(self.version, vrrpv2.version)
        self.assertEqual(self.type_, vrrpv2.type)
        self.assertEqual(self.vrid, vrrpv2.vrid)
        self.assertEqual(self.priority, vrrpv2.priority)
        self.assertEqual(self.count_ip, vrrpv2.count_ip)
        self.assertEqual(self.auth_type, vrrpv2.auth_type)
        self.assertEqual(self.max_adver_int, vrrpv2.max_adver_int)
        self.assertEqual(self.checksum, vrrpv2.checksum)
        self.assertEqual(1, len(vrrpv2.ip_addresses))
        self.assertEqual(str, type(vrrpv2.ip_addresses[0]))
        self.assertEqual(self.ip_address, vrrpv2.ip_addresses[0])
        self.assertEqual(self.auth_data, vrrpv2.auth_data)

    def test_serialize(self):
        src_ip = '192.168.0.1'
        dst_ip = vrrp.VRRP_IPV4_DST_ADDRESS
        prev = ipv4.ipv4(4, 5, 0, 0, 0, 0, 0, vrrp.VRRP_IPV4_TTL,
                         inet.IPPROTO_VRRP, 0, src_ip, dst_ip)

        type_ = vrrp.VRRP_TYPE_ADVERTISEMENT
        vrid = 5
        priority = 10
        max_adver_int = 30
        ip_address = '192.168.0.2'
        ip_addresses = [ip_address]

        vrrp_ = vrrp.vrrpv2.create(
            type_, vrid, priority, max_adver_int, ip_addresses)

        buf = vrrp_.serialize(bytearray(), prev)
        pack_str = vrrp.vrrpv2._PACK_STR + '4sII'
        pack_len = struct.calcsize(pack_str)
        res = struct.unpack(pack_str, bytes(buf))
        self.assertEqual(res[0], vrrp.vrrp_to_version_type(vrrp.VRRP_VERSION_V2, type_))
        self.assertEqual(res[1], vrid)
        self.assertEqual(res[2], priority)
        self.assertEqual(res[3], len(ip_addresses))
        self.assertEqual(res[4], vrrp.VRRP_AUTH_NO_AUTH)
        self.assertEqual(res[5], max_adver_int)
        # res[6] is checksum
        self.assertEqual(res[7], addrconv.ipv4.text_to_bin(ip_address))
        self.assertEqual(res[8], 0)
        self.assertEqual(res[9], 0)
        self.assertEqual(len(buf), pack_len)

        # checksum
        s = packet_utils.checksum(buf)
        self.assertEqual(0, s)

    def test_malformed_vrrpv2(self):
        m_short_buf = self.buf[1:vrrp.vrrpv2._MIN_LEN]
        self.assertRaises(Exception, vrrp.vrrp.parser, m_short_buf)

    def test_create_packet(self):
        primary_ip = '192.168.0.2'
        p0 = self.vrrpv2.create_packet(primary_ip)
        p0.serialize()
        p1 = packet.Packet(bytes(p0.data))
        p1.serialize()
        self.assertEqual(p0.data, p1.data)

    def _test_is_valid(self, type_=None, vrid=None, priority=None,
                       max_adver_int=None):
        if type_ is None:
            type_ = self.type_
        if vrid is None:
            vrid = self.vrid
        if priority is None:
            priority = self.priority
        if max_adver_int is None:
            max_adver_int = self.max_adver_int

        vrrp_ = vrrp.vrrpv2.create(type_, vrid, priority, max_adver_int,
                                   [self.ip_address])
        return vrrp_.is_valid()

    def test_is_valid_ok(self):
        self.assertTrue(self._test_is_valid())

    def test_is_valid_ng_type(self):
        self.assertTrue(not self._test_is_valid(type_=15))

    def test_is_valid_ng_vrid_min(self):
        vrid = vrrp.VRRP_VRID_MIN - 1
        self.assertTrue(not self._test_is_valid(vrid=vrid))

    def test_is_valid_ng_vrid_max(self):
        vrid = vrrp.VRRP_VRID_MAX + 1
        self.assertTrue(not self._test_is_valid(vrid=vrid))

    def test_is_valid_ng_priority_min(self):
        priority = vrrp.VRRP_PRIORITY_MIN - 1
        self.assertTrue(not self._test_is_valid(priority=priority))

    def test_is_valid_ng_priority_max(self):
        priority = vrrp.VRRP_PRIORITY_MAX + 1
        self.assertTrue(not self._test_is_valid(priority=priority))

    def test_is_valid_ng_adver_min(self):
        max_adver_int = vrrp.VRRP_V2_MAX_ADVER_INT_MIN - 1
        self.assertTrue(not self._test_is_valid(max_adver_int=max_adver_int))

    def test_is_valid_ng_adver_max(self):
        max_adver_int = vrrp.VRRP_V2_MAX_ADVER_INT_MAX + 1
        self.assertTrue(not self._test_is_valid(max_adver_int=max_adver_int))

    def test_to_string(self):
        vrrpv2_values = {'version': self.version,
                         'type': self.type_,
                         'vrid': self.vrid,
                         'priority': self.priority,
                         'count_ip': self.count_ip,
                         'max_adver_int': self.max_adver_int,
                         'checksum': self.vrrpv2.checksum,
                         'ip_addresses': [self.ip_address],
                         'auth_type': self.auth_type,
                         'auth_data': self.auth_data,
                         'identification': self.vrrpv2.identification}
        _vrrpv2_str = ','.join(['%s=%s' % (k, repr(vrrpv2_values[k]))
                                for k, v in inspect.getmembers(self.vrrpv2)
                                if k in vrrpv2_values])
        vrrpv2_str = '%s(%s)' % (vrrp.vrrpv2.__name__, _vrrpv2_str)

        self.assertEqual(str(self.vrrpv2), vrrpv2_str)
        self.assertEqual(repr(self.vrrpv2), vrrpv2_str)


class Test_vrrpv3_ipv4(unittest.TestCase):
    """ Test case for vrrp v3 IPv4
    """
    version = vrrp.VRRP_VERSION_V3
    type_ = vrrp.VRRP_TYPE_ADVERTISEMENT
    vrid = 128
    priority = 99
    count_ip = 1
    max_adver_int = 111
    checksum = 0
    ip_address = '192.168.0.1'
    vrrpv3 = vrrp.vrrpv3.create(type_, vrid, priority, max_adver_int,
                                [ip_address])
    buf = struct.pack(vrrp.vrrpv3._PACK_STR + '4s',
                      vrrp.vrrp_to_version_type(vrrp.VRRP_VERSION_V3, type_),
                      vrid, priority, count_ip,
                      max_adver_int, checksum,
                      addrconv.ipv4.text_to_bin(ip_address))

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.type_, self.vrrpv3.type)
        self.assertEqual(self.vrid, self.vrrpv3.vrid)
        self.assertEqual(self.priority, self.vrrpv3.priority)
        self.assertEqual(self.count_ip, self.vrrpv3.count_ip)
        self.assertEqual(1, len(self.vrrpv3.ip_addresses))
        self.assertEqual(self.ip_address, self.vrrpv3.ip_addresses[0])

    def test_parser(self):
        vrrpv3, _cls, _ = self.vrrpv3.parser(self.buf)

        self.assertEqual(self.version, vrrpv3.version)
        self.assertEqual(self.type_, vrrpv3.type)
        self.assertEqual(self.vrid, vrrpv3.vrid)
        self.assertEqual(self.priority, vrrpv3.priority)
        self.assertEqual(self.count_ip, vrrpv3.count_ip)
        self.assertEqual(self.max_adver_int, vrrpv3.max_adver_int)
        self.assertEqual(self.checksum, vrrpv3.checksum)
        self.assertEqual(1, len(vrrpv3.ip_addresses))
        self.assertEqual(str, type(vrrpv3.ip_addresses[0]))
        self.assertEqual(self.ip_address, vrrpv3.ip_addresses[0])

    def test_serialize(self):
        src_ip = '192.168.0.1'
        dst_ip = vrrp.VRRP_IPV4_DST_ADDRESS
        prev = ipv4.ipv4(4, 5, 0, 0, 0, 0, 0, vrrp.VRRP_IPV4_TTL,
                         inet.IPPROTO_VRRP, 0, src_ip, dst_ip)

        type_ = vrrp.VRRP_TYPE_ADVERTISEMENT
        vrid = 5
        priority = 10
        max_adver_int = 30
        ip_address = '192.168.0.2'
        ip_addresses = [ip_address]

        vrrp_ = vrrp.vrrpv3.create(
            type_, vrid, priority, max_adver_int, ip_addresses)

        buf = vrrp_.serialize(bytearray(), prev)
        print(len(buf), type(buf), buf)
        pack_str = vrrp.vrrpv3._PACK_STR + '4s'
        pack_len = struct.calcsize(pack_str)
        res = struct.unpack(pack_str, bytes(buf))
        self.assertEqual(res[0], vrrp.vrrp_to_version_type(vrrp.VRRP_VERSION_V3, type_))
        self.assertEqual(res[1], vrid)
        self.assertEqual(res[2], priority)
        self.assertEqual(res[3], len(ip_addresses))
        self.assertEqual(res[4], max_adver_int)
        # res[5] is checksum
        self.assertEqual(res[6], addrconv.ipv4.text_to_bin(ip_address))
        self.assertEqual(len(buf), pack_len)
        print(res)

        # checksum
        ph = struct.pack('!4s4sxBH',
                         addrconv.ipv4.text_to_bin(src_ip),
                         addrconv.ipv4.text_to_bin(dst_ip),
                         inet.IPPROTO_VRRP, pack_len)
        s = packet_utils.checksum(ph + buf)
        self.assertEqual(0, s)

    def test_malformed_vrrpv3(self):
        m_short_buf = self.buf[1:vrrp.vrrpv3._MIN_LEN]
        self.assertRaises(Exception, vrrp.vrrp.parser, m_short_buf)

    def test_create_packet(self):
        primary_ip = '192.168.0.2'
        p0 = self.vrrpv3.create_packet(primary_ip)
        p0.serialize()
        p1 = packet.Packet(bytes(p0.data))
        p1.serialize()
        self.assertEqual(p0.data, p1.data)

    def _test_is_valid(self, type_=None, vrid=None, priority=None,
                       max_adver_int=None):
        if type_ is None:
            type_ = self.type_
        if vrid is None:
            vrid = self.vrid
        if priority is None:
            priority = self.priority
        if max_adver_int is None:
            max_adver_int = self.max_adver_int

        vrrp_ = vrrp.vrrpv3.create(type_, vrid, priority, max_adver_int,
                                   [self.ip_address])
        return vrrp_.is_valid()

    def test_is_valid_ok(self):
        self.assertTrue(self._test_is_valid())

    def test_is_valid_ng_type(self):
        self.assertTrue(not self._test_is_valid(type_=15))

    def test_is_valid_ng_vrid_min(self):
        vrid = vrrp.VRRP_VRID_MIN - 1
        self.assertTrue(not self._test_is_valid(vrid=vrid))

    def test_is_valid_ng_vrid_max(self):
        vrid = vrrp.VRRP_VRID_MAX + 1
        self.assertTrue(not self._test_is_valid(vrid=vrid))

    def test_is_valid_ng_priority_min(self):
        priority = vrrp.VRRP_PRIORITY_MIN - 1
        self.assertTrue(not self._test_is_valid(priority=priority))

    def test_is_valid_ng_priority_max(self):
        priority = vrrp.VRRP_PRIORITY_MAX + 1
        self.assertTrue(not self._test_is_valid(priority=priority))

    def test_is_valid_ng_adver_min(self):
        max_adver_int = vrrp.VRRP_V3_MAX_ADVER_INT_MIN - 1
        self.assertTrue(not self._test_is_valid(max_adver_int=max_adver_int))

    def test_is_valid_ng_adver_max(self):
        max_adver_int = vrrp.VRRP_V3_MAX_ADVER_INT_MAX + 1
        self.assertTrue(not self._test_is_valid(max_adver_int=max_adver_int))

    def test_to_string(self):
        vrrpv3_values = {'version': self.version,
                         'type': self.type_,
                         'vrid': self.vrid,
                         'priority': self.priority,
                         'count_ip': self.count_ip,
                         'max_adver_int': self.max_adver_int,
                         'checksum': self.vrrpv3.checksum,
                         'ip_addresses': [self.ip_address],
                         'auth_type': None,
                         'auth_data': None,
                         'identification': self.vrrpv3.identification}
        _vrrpv3_str = ','.join(['%s=%s' % (k, repr(vrrpv3_values[k]))
                                for k, v in inspect.getmembers(self.vrrpv3)
                                if k in vrrpv3_values])
        vrrpv3_str = '%s(%s)' % (vrrp.vrrpv3.__name__, _vrrpv3_str)

        self.assertEqual(str(self.vrrpv3), vrrpv3_str)
        self.assertEqual(repr(self.vrrpv3), vrrpv3_str)


class Test_vrrpv3_ipv6(unittest.TestCase):
    """ Test case for vrrp v3 IPv6
    """
    version = vrrp.VRRP_VERSION_V3
    type_ = vrrp.VRRP_TYPE_ADVERTISEMENT
    vrid = 128
    priority = 99
    count_ip = 1
    max_adver_int = 111
    checksum = 0
    ip_address = '2001:db8:2000::1'
    vrrpv3 = vrrp.vrrpv3.create(type_, vrid, priority, max_adver_int,
                                [ip_address])
    buf = struct.pack(vrrp.vrrpv3._PACK_STR + '16s',
                      vrrp.vrrp_to_version_type(vrrp.VRRP_VERSION_V3, type_),
                      vrid, priority, count_ip,
                      max_adver_int, checksum,
                      addrconv.ipv6.text_to_bin(ip_address))

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_init(self):
        self.assertEqual(self.type_, self.vrrpv3.type)
        self.assertEqual(self.vrid, self.vrrpv3.vrid)
        self.assertEqual(self.priority, self.vrrpv3.priority)
        self.assertEqual(self.count_ip, self.vrrpv3.count_ip)
        self.assertEqual(1, len(self.vrrpv3.ip_addresses))
        self.assertEqual(self.ip_address, self.vrrpv3.ip_addresses[0])

    def test_parser(self):
        vrrpv3, _cls, _ = self.vrrpv3.parser(self.buf)

        self.assertEqual(self.version, vrrpv3.version)
        self.assertEqual(self.type_, vrrpv3.type)
        self.assertEqual(self.vrid, vrrpv3.vrid)
        self.assertEqual(self.priority, vrrpv3.priority)
        self.assertEqual(self.count_ip, vrrpv3.count_ip)
        self.assertEqual(self.max_adver_int, vrrpv3.max_adver_int)
        self.assertEqual(self.checksum, vrrpv3.checksum)
        self.assertEqual(1, len(vrrpv3.ip_addresses))
        self.assertEqual(str, type(vrrpv3.ip_addresses[0]))
        self.assertEqual(self.ip_address, vrrpv3.ip_addresses[0])

    def test_serialize(self):
        src_ip = '2001:db8:2000::1'
        dst_ip = vrrp.VRRP_IPV6_DST_ADDRESS
        prev = ipv6.ipv6(6, 0, 0, 0, inet.IPPROTO_VRRP,
                         vrrp.VRRP_IPV6_HOP_LIMIT, src_ip, dst_ip)

        type_ = vrrp.VRRP_TYPE_ADVERTISEMENT
        vrid = 5
        priority = 10
        max_adver_int = 30
        ip_address = '2001:db8:2000::2'
        ip_addresses = [ip_address]

        vrrp_ = vrrp.vrrpv3.create(
            type_, vrid, priority, max_adver_int, ip_addresses)

        buf = vrrp_.serialize(bytearray(), prev)
        print(len(buf), type(buf), buf)
        pack_str = vrrp.vrrpv3._PACK_STR + '16s'
        pack_len = struct.calcsize(pack_str)
        res = struct.unpack(pack_str, bytes(buf))
        self.assertEqual(res[0], vrrp.vrrp_to_version_type(vrrp.VRRP_VERSION_V3, type_))
        self.assertEqual(res[1], vrid)
        self.assertEqual(res[2], priority)
        self.assertEqual(res[3], len(ip_addresses))
        self.assertEqual(res[4], max_adver_int)
        # res[5] is checksum
        self.assertEqual(res[6], addrconv.ipv6.text_to_bin(ip_address))
        self.assertEqual(len(buf), pack_len)
        print(res)

        # checksum
        ph = struct.pack('!16s16sI3xB',
                         addrconv.ipv6.text_to_bin(src_ip),
                         addrconv.ipv6.text_to_bin(dst_ip),
                         pack_len, inet.IPPROTO_VRRP)
        s = packet_utils.checksum(ph + buf)
        self.assertEqual(0, s)

    def test_malformed_vrrpv3(self):
        m_short_buf = self.buf[1:vrrp.vrrpv3._MIN_LEN]
        self.assertRaises(Exception, vrrp.vrrp.parser, m_short_buf)

    def test_create_packet(self):
        primary_ip = '2001:db8:2000::3'
        p0 = self.vrrpv3.create_packet(primary_ip)
        p0.serialize()
        print(len(p0.data), p0.data)
        p1 = packet.Packet(bytes(p0.data))
        p1.serialize()
        print(len(p0.data), p0.data)
        print(len(p1.data), p1.data)
        self.assertEqual(p0.data, p1.data)

    def test_to_string(self):
        vrrpv3_values = {'version': self.version,
                         'type': self.type_,
                         'vrid': self.vrid,
                         'priority': self.priority,
                         'count_ip': self.count_ip,
                         'max_adver_int': self.max_adver_int,
                         'checksum': self.vrrpv3.checksum,
                         'ip_addresses': [self.ip_address],
                         'auth_type': None,
                         'auth_data': None,
                         'identification': self.vrrpv3.identification}
        _vrrpv3_str = ','.join(['%s=%s' % (k, repr(vrrpv3_values[k]))
                                for k, v in inspect.getmembers(self.vrrpv3)
                                if k in vrrpv3_values])
        vrrpv3_str = '%s(%s)' % (vrrp.vrrpv3.__name__, _vrrpv3_str)

        self.assertEqual(str(self.vrrpv3), vrrpv3_str)
        self.assertEqual(repr(self.vrrpv3), vrrpv3_str)



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2013 YAMAMOTO Takashi <yamamoto at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import itertools
import unittest
import testscenarios

from os_ken.ofproto import ofproto_v1_2
from os_ken.ofproto import ofproto_v1_3
from os_ken.ofproto import ofproto_v1_4
from os_ken.ofproto import ofproto_v1_5
from os_ken.ofproto import ofproto_v1_2_parser
from os_ken.ofproto import ofproto_v1_3_parser
from os_ken.ofproto import ofproto_v1_4_parser
from os_ken.ofproto import ofproto_v1_5_parser


def _list_test_cases():

    class Field(object):
        @classmethod
        def generate_mask(cls):
            return list(cls.generate())[1]

    class Int1(Field):
        @staticmethod
        def generate():
            yield 0
            yield 0xff

    class Int2(Field):
        @staticmethod
        def generate():
            yield 0
            yield 0x1234
            yield 0xffff

    class Int3(Field):
        @staticmethod
        def generate():
            yield 0
            yield 0x123456
            yield 0xffffff

    class Int4(Field):
        @staticmethod
        def generate():
            yield 0
            yield 0x12345678
            yield 0xffffffff

    class Int8(Field):
        @staticmethod
        def generate():
            yield 0
            yield 0x123456789abcdef0
            yield 0xffffffffffffffff

    class Mac(Field):
        @staticmethod
        def generate():
            yield '00:00:00:00:00:00'
            yield 'f2:0b:a4:7d:f8:ea'
            yield 'ff:ff:ff:ff:ff:ff'

    class IPv4(Field):
        @staticmethod
        def generate():
            yield '0.0.0.0'
            yield '192.0.2.1'
            yield '255.255.255.255'

    class IPv6(Field):
        @staticmethod
        def generate():
            yield '::'
            yield 'fe80::f00b:a4ff:fed0:3f70'
            yield 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'

    class B64(Field):
        @staticmethod
        def generate():
            yield 'aG9nZWhvZ2U='
            yield 'ZnVnYWZ1Z2E='

    ofpps = [ofproto_v1_2_parser, ofproto_v1_3_parser,
             ofproto_v1_4_parser, ofproto_v1_5_parser]
    common = [
        # OpenFlow Basic
        ('in_port', Int4),
        ('in_phy_port', Int4),
        ('metadata', Int8),
        ('eth_dst', Mac),
        ('eth_src', Mac),
        ('eth_type', Int2),
        ('vlan_vid', Int2),
        ('vlan_pcp', Int1),
        ('ip_dscp', Int1),
        ('ip_ecn', Int1),
        ('ip_proto', Int1),
        ('ipv4_src', IPv4),
        ('ipv4_dst', IPv4),
        ('tcp_src', Int2),
        ('tcp_dst', Int2),
        ('udp_src', Int2),
        ('udp_dst', Int2),
        ('sctp_src', Int2),
        ('sctp_dst', Int2),
        ('icmpv4_type', Int1),
        ('icmpv4_code', Int1),
        ('arp_op', Int2),
        ('arp_spa', IPv4),
        ('arp_tpa', IPv4),
        ('arp_sha', Mac),
        ('arp_tha', Mac),
        ('ipv6_src', IPv6),
        ('ipv6_dst', IPv6),
        ('ipv6_flabel', Int4),
        ('icmpv6_type', Int1),
        ('icmpv6_code', Int1),
        ('ipv6_nd_target', IPv6),
        ('ipv6_nd_sll', Mac),
        ('ipv6_nd_tll', Mac),
        ('mpls_label', Int4),
        ('mpls_tc', Int1),
        # Old ONF Experimenter --> OpenFlow Basic (OF1.4+)
        ('pbb_uca', Int1),
        # ONF Experimenter --> OpenFlow Basic (OF1.5+)
        ('tcp_flags', Int2),
        ('actset_output', Int4),
        # Nicira Experimenter
        ('eth_dst_nxm', Mac),
        ('eth_src_nxm', Mac),
        ('tunnel_id_nxm', Int8),
        ('tun_ipv4_src', IPv4),
        ('tun_ipv4_dst', IPv4),
        ('pkt_mark', Int4),
        ('conj_id', Int4),
        ('tun_ipv6_src', IPv6),
        ('tun_ipv6_dst', IPv6),
        ('_dp_hash', Int4),
        ('reg0', Int4),
        ('reg1', Int4),
        ('reg2', Int4),
        ('reg3', Int4),
        ('reg4', Int4),
        ('reg5', Int4),
        ('reg6', Int4),
        ('reg7', Int4),
        # Common Experimenter
        ('field_100', B64),
    ]
    L = {}
    L[ofproto_v1_2_parser] = common + [
        # OF1.2 doesn't have OXM_OF_PBB_ISID.
        #    OFPXMC_OPENFLOW_BASIC = 0x8000
        #    OXM_OF_PBB_ISID = 37
        #    (OFPXMC_OPENFLOW_BASIC << 7) + OXM_OF_PBB_ISID == 4194341
        ('field_4194341', B64),
    ]
    L[ofproto_v1_3_parser] = common + [
        # OpenFlow Basic (OF1.3+)
        ('mpls_bos', Int1),
        ('pbb_isid', Int3),
        ('tunnel_id', Int8),
        ('ipv6_exthdr', Int2),
    ]
    L[ofproto_v1_4_parser] = L[ofproto_v1_3_parser]
    L[ofproto_v1_5_parser] = L[ofproto_v1_4_parser] + [
        # OpenFlow Basic (OF1.5+)
        ('packet_type', Int4),
    ]

    def flatten_one(l, i):
        if isinstance(i, tuple):
            return l + flatten(i)
        else:
            return l + [i]
    flatten = lambda l: functools.reduce(flatten_one, l, [])

    cases = []
    for ofpp in ofpps:
        for n in range(1, 3):
            for C in itertools.combinations(L[ofpp], n):
                l = [1]
                keys = []
                clss = []
                for (k, cls) in C:
                    l = itertools.product(l, cls.generate())
                    keys.append(k)
                    clss.append(cls)
                l = [flatten(x)[1:] for x in l]
                for domask in [True, False]:
                    for values in l:
                        if domask:
                            values = [(value, cls.generate_mask())
                                      for (cls, value)
                                      in zip(clss, values)]
                        d = dict(zip(keys, values))
                        mod = ofpp.__name__.split('.')[-1]
                        method_name = 'test_' + mod
                        if domask:
                            method_name += '_mask'
                        for k in sorted(dict(d).keys()):
                            method_name += '_' + str(k)
                            method_name += '_' + str(d[k])
                        method_name = method_name.replace(':', '_')
                        method_name = method_name.replace('.', '_')
                        method_name = method_name.replace('(', '_')
                        method_name = method_name.replace(')', '_')
                        method_name = method_name.replace(',', '_')
                        method_name = method_name.replace("'", '_')
                        method_name = method_name.replace(' ', '_')
                        cases.append({'name': method_name, 'ofpp': ofpp,
                                      'd': d, 'domask': domask})
    return cases


class Test_Parser_OFPMatch(testscenarios.WithScenarios, unittest.TestCase):

    scenarios = [(case['name'], case) for case in _list_test_cases()]

    _ofp = {ofproto_v1_2_parser: ofproto_v1_2,
            ofproto_v1_3_parser: ofproto_v1_3,
            ofproto_v1_4_parser: ofproto_v1_4,
            ofproto_v1_5_parser: ofproto_v1_5}

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_parser(self):
        self._test(name=self.name, ofpp=self.ofpp, d=self.d,
                   domask=self.domask)

    def _test(self, name, ofpp, d, domask):
        if domask:
            d = dict(self._ofp[ofpp].oxm_normalize_user(k, uv)
                     for (k, uv)
                     in d.items())
        match = ofpp.OFPMatch(**d)
        b = bytearray()
        match.serialize(b, 0)
        match2 = match.parser(bytes(b), 0)
        for k, v in d.items():
            self.assertTrue(k in match)
            self.assertTrue(k in match2)
            self.assertEqual(match[k], v)
            self.assertEqual(match2[k], v)
        for k, v in match.iteritems():
            self.assertTrue(k in d)
            self.assertEqual(d[k], v)
        for k, v in match2.iteritems():
            self.assertTrue(k in d)
            self.assertEqual(d[k], v)



# Copyright (C) 2011-2014 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2011, 2012 Isaku Yamahata <yamahata at valinux co jp>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import logging
import sys

LOG = logging.getLogger('os_ken.controller.handler')

# just represent OF datapath state. datapath specific so should be moved.
HANDSHAKE_DISPATCHER = "handshake"
CONFIG_DISPATCHER = "config"
MAIN_DISPATCHER = "main"
DEAD_DISPATCHER = "dead"


class _Caller(object):
    """Describe how to handle an event class.
    """

    def __init__(self, dispatchers, ev_source):
        """Initialize _Caller.

        :param dispatchers: A list of states or a state, in which this
                            is in effect.
                            None and [] mean all states.
        :param ev_source: The module which generates the event.
                          ev_cls.__module__ for set_ev_cls.
                          None for set_ev_handler.
        """
        self.dispatchers = dispatchers
        self.ev_source = ev_source


# should be named something like 'observe_event'
def set_ev_cls(ev_cls, dispatchers=None):
    """
    A decorator for OSKen application to declare an event handler.

    Decorated method will become an event handler.
    ev_cls is an event class whose instances this OSKenApp wants to receive.
    dispatchers argument specifies one of the following negotiation phases
    (or a list of them) for which events should be generated for this handler.
    Note that, in case an event changes the phase, the phase before the change
    is used to check the interest.

    .. tabularcolumns:: |l|L|

    ============================================== ===============================
    Negotiation phase                              Description
    ============================================== ===============================
    os_ken.controller.handler.HANDSHAKE_DISPATCHER Sending and waiting for hello
                                                   message
    os_ken.controller.handler.CONFIG_DISPATCHER    Version negotiated and sent
                                                   features-request message
    os_ken.controller.handler.MAIN_DISPATCHER      Switch-features message
                                                   received and sent set-config
                                                   message
    os_ken.controller.handler.DEAD_DISPATCHER      Disconnect from the peer.  Or
                                                   disconnecting due to some
                                                   unrecoverable errors.
    ============================================== ===============================
    """
    def _set_ev_cls_dec(handler):
        if 'callers' not in dir(handler):
            handler.callers = {}
        for e in _listify(ev_cls):
            handler.callers[e] = _Caller(_listify(dispatchers), e.__module__)
        return handler
    return _set_ev_cls_dec


def set_ev_handler(ev_cls, dispatchers=None):
    def _set_ev_cls_dec(handler):
        if 'callers' not in dir(handler):
            handler.callers = {}
        for e in _listify(ev_cls):
            handler.callers[e] = _Caller(_listify(dispatchers), None)
        return handler
    return _set_ev_cls_dec


def _has_caller(meth):
    return hasattr(meth, 'callers')


def _listify(may_list):
    if may_list is None:
        may_list = []
    if not isinstance(may_list, list):
        may_list = [may_list]
    return may_list


def register_instance(i):
    for _k, m in inspect.getmembers(i, inspect.ismethod):
        # LOG.debug('instance %s k %s m %s', i, _k, m)
        if _has_caller(m):
            for ev_cls, c in m.callers.items():
                i.register_handler(ev_cls, m)


def _is_method(f):
    return inspect.isfunction(f) or inspect.ismethod(f)


def get_dependent_services(cls):
    services = []
    for _k, m in inspect.getmembers(cls, _is_method):
        if _has_caller(m):
            for ev_cls, c in m.callers.items():
                service = getattr(sys.modules[ev_cls.__module__],
                                  '_SERVICE_NAME', None)
                if service:
                    # avoid cls that registers the own events (like
                    # ofp_handler)
                    if cls.__module__ != service:
                        services.append(service)

    m = sys.modules[cls.__module__]
    services.extend(getattr(m, '_REQUIRED_APP', []))
    services = list(set(services))
    return services


def register_service(service):
    """
    Register the os_ken application specified by 'service' as
    a provider of events defined in the calling module.

    If an application being loaded consumes events (in the sense of
    set_ev_cls) provided by the 'service' application, the latter
    application will be automatically loaded.

    This mechanism is used to e.g. automatically start ofp_handler if
    there are applications consuming OFP events.
    """
    frame = inspect.currentframe()
    if frame is not None:
        m_name = frame.f_back.f_globals['__name__']
        m = sys.modules[m_name]
        m._SERVICE_NAME = service



# Copyright (c) 2014 Rackspace Hosting
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import errno
import uuid

from ovs import jsonrpc
from ovs import poller
from ovs import reconnect
from ovs import stream
from ovs import timeval
from ovs.db import idl

from os_ken.base import app_manager
from os_ken.lib import hub
from os_ken.services.protocols.ovsdb import event
from os_ken.services.protocols.ovsdb import model


now = timeval.msec


def _uuid_to_row(atom, base):
    if base.ref_table:
        value = base.ref_table.rows.get(atom)
    else:
        value = atom

    if isinstance(value, idl.Row):
        value = str(value.uuid)

    return value


def dictify(row):
    if row is None:
        return

    result = {}

    for key, value in row._data.items():
        result[key] = value.to_python(_uuid_to_row)
        hub.sleep(0)

    return result


def transact_block(request, connection):
    """Emulate jsonrpc.Connection.transact_block without blocking eventlet.
    """
    error = connection.send(request)
    reply = None

    if error:
        return error, reply

    ovs_poller = poller.Poller()
    while not error:
        ovs_poller.immediate_wake()
        error, reply = connection.recv()

        if error != errno.EAGAIN:
            break

        if (reply and
            reply.id == request.id and
            reply.type in (jsonrpc.Message.T_REPLY,
                           jsonrpc.Message.T_ERROR)):
            break

        connection.run()
        connection.wait(ovs_poller)
        connection.recv_wait(ovs_poller)
        ovs_poller.block()

        hub.sleep(0)

    return error, reply


def discover_schemas(connection):
    # NOTE(jkoelker) currently only the Open_vSwitch schema
    #                is supported.
    # TODO(jkoelker) support arbitrary schemas
    req = jsonrpc.Message.create_request('list_dbs', [])
    error, reply = transact_block(req, connection)

    if error or reply.error:
        return

    schemas = []
    for db in reply.result:
        if db != 'Open_vSwitch':
            continue

        req = jsonrpc.Message.create_request('get_schema', [db])
        error, reply = transact_block(req, connection)

        if error or reply.error:
            # TODO(jkoelker) Error handling
            continue

        schemas.append(reply.result)

    return schemas


def discover_system_id(idl):
    system_id = None

    while system_id is None and idl._session.is_connected():
        idl.run()
        openvswitch = idl.tables['Open_vSwitch'].rows

        if openvswitch:
            row = openvswitch.get(list(openvswitch.keys())[0])
            system_id = row.external_ids.get('system-id')

    return system_id


def _filter_schemas(schemas, schema_tables, exclude_table_columns):
    """Wrapper method for _filter_schema to filter multiple schemas."""
    return [_filter_schema(s, schema_tables, exclude_table_columns)
            for s in schemas]


def _filter_schema(schema, schema_tables, exclude_table_columns):
    """Filters a schema to only include the specified tables in the
       schema_tables parameter.  This will also filter out any colums for
       included tables that reference tables that are not included
       in the schema_tables parameter

    :param schema: Schema dict to be filtered
    :param schema_tables: List of table names to filter on.
                          EX: ['Bridge', 'Controller', 'Interface']
                          NOTE: This list is case sensitive.
    :return: Schema dict:
                filtered if the schema_table parameter contains table names,
                else the original schema dict
    """

    tables = {}
    for tbl_name, tbl_data in schema['tables'].items():
        if not schema_tables or tbl_name in schema_tables:
            columns = {}

            exclude_columns = exclude_table_columns.get(tbl_name, [])
            for col_name, col_data in tbl_data['columns'].items():
                if col_name in exclude_columns:
                    continue

                # NOTE(Alan Quillin) Needs to check and remove
                # and columns that have references to tables that
                # are not to be configured
                type_ = col_data.get('type')
                if type_:
                    if type_ and isinstance(type_, dict):
                        key = type_.get('key')
                        if key and isinstance(key, dict):
                            ref_tbl = key.get('refTable')
                            if ref_tbl and isinstance(ref_tbl,
                                                      str):
                                if ref_tbl not in schema_tables:
                                    continue
                        value = type_.get('value')
                        if value and isinstance(value, dict):
                            ref_tbl = value.get('refTable')
                            if ref_tbl and isinstance(ref_tbl,
                                                      str):
                                if ref_tbl not in schema_tables:
                                    continue

                columns[col_name] = col_data

            tbl_data['columns'] = columns
            tables[tbl_name] = tbl_data

    schema['tables'] = tables

    return schema


# NOTE(jkoelker) Wrap ovs's Idl to accept an existing session, and
#                trigger callbacks on changes
class Idl(idl.Idl):
    def __init__(self, session, schema):
        if not isinstance(schema, idl.SchemaHelper):
            schema = idl.SchemaHelper(schema_json=schema)
            schema.register_all()

        schema = schema.get_idl_schema()

        # NOTE(jkoelker) event buffer
        self._events = []

        self.tables = schema.tables
        self.readonly = schema.readonly
        self._db = schema
        self._session = session
        self._monitor_request_id = None
        self._last_seqno = None
        self.change_seqno = 0
        self.uuid = uuid.uuid1()
        self.state = self.IDL_S_INITIAL

        # Database locking.
        self.lock_name = None          # Name of lock we need, None if none.
        self.has_lock = False          # Has db server said we have the lock?
        self.is_lock_contended = False  # Has db server said we can't get lock?
        self._lock_request_id = None   # JSON-RPC ID of in-flight lock request.

        # Transaction support.
        self.txn = None
        self._outstanding_txns = {}

        for table in schema.tables.values():
            for column in table.columns.values():
                if not hasattr(column, 'alert'):
                    column.alert = True
            table.need_table = False
            table.rows = {}
            table.idl = self
            table.condition = []
            table.cond_changed = False

    @property
    def events(self):
        events = self._events
        self._events = []
        return events

    def __process_update(self, table, uuid, old, new):
        old_row = table.rows.get(uuid)
        if old_row is not None:
            old_row = model.Row(dictify(old_row))
            old_row['_uuid'] = uuid

        changed = idl.Idl.__process_update(self, table, uuid, old, new)

        if changed:
            if not new:
                ev = (event.EventRowDelete, (table.name, old_row))

            elif not old:
                new_row = model.Row(dictify(table.rows.get(uuid)))
                new_row['_uuid'] = uuid
                ev = (event.EventRowInsert, (table.name, new_row))

            else:
                new_row = model.Row(dictify(table.rows.get(uuid)))
                new_row['_uuid'] = uuid

                ev = (event.EventRowUpdate, (table.name, old_row, new_row))

            self._events.append(ev)

        return changed


class RemoteOvsdb(app_manager.OSKenApp):
    _EVENTS = [event.EventRowUpdate,
               event.EventRowDelete,
               event.EventRowInsert,
               event.EventInterfaceDeleted,
               event.EventInterfaceInserted,
               event.EventInterfaceUpdated,
               event.EventPortDeleted,
               event.EventPortInserted,
               event.EventPortUpdated]

    @classmethod
    def factory(cls, sock, address, probe_interval=None, min_backoff=None,
                max_backoff=None, schema_tables=None,
                schema_exclude_columns=None, *args, **kwargs):
        schema_exclude_columns = schema_exclude_columns or {}
        ovs_stream = stream.Stream(sock, None, None)
        connection = jsonrpc.Connection(ovs_stream)
        schemas = discover_schemas(connection)

        if not schemas:
            return

        if schema_tables or schema_exclude_columns:
            schemas = _filter_schemas(schemas, schema_tables,
                                      schema_exclude_columns)

        fsm = reconnect.Reconnect(now())
        fsm.set_name('%s:%s' % address[:2])
        fsm.enable(now())
        fsm.set_passive(True, now())
        fsm.set_max_tries(-1)

        if probe_interval is not None:
            fsm.set_probe_interval(probe_interval)

        if min_backoff is None:
            min_backoff = fsm.get_min_backoff()

        if max_backoff is None:
            max_backoff = fsm.get_max_backoff()

        if min_backoff and max_backoff:
            fsm.set_backoff(min_backoff, max_backoff)

        fsm.connected(now())

        session = jsonrpc.Session(fsm, connection, fsm.get_name())
        idl = Idl(session, schemas[0])

        system_id = discover_system_id(idl)

        if not system_id:
            return None

        name = cls.instance_name(system_id)
        ovs_stream.name = name
        connection.name = name
        fsm.set_name(name)

        kwargs = kwargs.copy()
        kwargs['socket'] = sock
        kwargs['address'] = address
        kwargs['idl'] = idl
        kwargs['name'] = name
        kwargs['system_id'] = system_id

        app_mgr = app_manager.AppManager.get_instance()

        old_app = app_manager.lookup_service_brick(name)
        old_events = None
        if old_app:
            old_events = old_app.events
            app_mgr.uninstantiate(name)

        app = app_mgr.instantiate(cls, *args, **kwargs)

        if old_events:
            app.events = old_events

        return app

    @classmethod
    def instance_name(cls, system_id):
        return '%s-%s' % (cls.__name__, system_id)

    def __init__(self, *args, **kwargs):
        super(RemoteOvsdb, self).__init__(*args, **kwargs)
        self.socket = kwargs['socket']
        self.address = kwargs['address']
        self._idl = kwargs['idl']
        self.system_id = kwargs['system_id']
        self.name = kwargs['name']
        self._txn_q = collections.deque()

    def _event_proxy_loop(self):
        while self.is_active:
            events = self._idl.events

            if not events:
                hub.sleep(0.1)
                continue

            for e in events:
                ev = e[0]
                args = e[1]
                self._submit_event(ev(self.system_id, *args))

            hub.sleep(0)

    def _submit_event(self, ev):
        self.send_event_to_observers(ev)
        try:
            ev_cls_name = 'Event' + ev.table + ev.event_type
            proxy_ev_cls = getattr(event, ev_cls_name, None)
            if proxy_ev_cls:
                self.send_event_to_observers(proxy_ev_cls(ev))
        except Exception:
            self.logger.exception(
                'Error submitting specific event for OVSDB %s', self.system_id)

    def _idl_loop(self):
        while self.is_active:
            try:
                self._idl.run()
                self._transactions()
            except Exception:
                self.logger.exception('Error running IDL for system_id %s' %
                                      self.system_id)
                raise

            hub.sleep(0)

    def _run_thread(self, func, *args, **kwargs):
        try:
            func(*args, **kwargs)

        except:
            self.stop()

    def _transactions(self):
        if not self._txn_q:
            return

        # NOTE(jkoelker) possibly run multiple transactions per loop?
        self._transaction()

    def _transaction(self):
        req = self._txn_q.popleft()
        txn = idl.Transaction(self._idl)

        uuids = req.func(self._idl.tables, txn.insert)
        status = txn.commit_block()

        insert_uuids = {}
        err_msg = None

        if status in (idl.Transaction.SUCCESS,
                      idl.Transaction.UNCHANGED):
            if uuids:
                if isinstance(uuids, uuid.UUID):
                    insert_uuids[uuids] = txn.get_insert_uuid(uuids)

                else:
                    insert_uuids = dict((uuid, txn.get_insert_uuid(uuid))
                                        for uuid in uuids)
        else:
            err_msg = txn.get_error()

        rep = event.EventModifyReply(self.system_id, status, insert_uuids,
                                     err_msg)
        self.reply_to_request(req, rep)

    def modify_request_handler(self, ev):
        self._txn_q.append(ev)

    def read_request_handler(self, ev, bulk=False):
        result = ev.func(self._idl.tables)

        # NOTE(jkoelker) If this was a bulk request, the parent OVSDB app is
        #                responsible for the reply

        if bulk:
            return (self.system_id, result)

        rep = event.EventReadReply(self.system_id, result)
        self.reply_to_request(ev, rep)

    def start(self):
        super(RemoteOvsdb, self).start()
        t = hub.spawn(self._run_thread, self._idl_loop)
        self.threads.append(t)

        t = hub.spawn(self._run_thread, self._event_proxy_loop)
        self.threads.append(t)

    def stop(self):
        # NOTE(jkoelker) Stop the idl and event_proxy threads first
        #                letting them finish their current loop.
        self.is_active = False
        hub.joinall(self.threads)

        self._idl.close()
        super(RemoteOvsdb, self).stop()



# Copyright (C) 2014 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os_ken.services.protocols.bgp.info_base.vrf import VrfRtImportMap
from os_ken.services.protocols.bgp.info_base.vrf4 import Vrf4NlriImportMap
from os_ken.services.protocols.bgp.info_base.vrf6 import Vrf6NlriImportMap


class ImportMapManager(object):

    def __init__(self):
        self._import_maps_by_name = {}

    def create_vpnv4_nlri_import_map(self, name, value):
        self._create_import_map_factory(name, value, Vrf4NlriImportMap)

    def create_vpnv6_nlri_import_map(self, name, value):
        self._create_import_map_factory(name, value, Vrf6NlriImportMap)

    def create_rt_import_map(self, name, value):
        self._create_import_map_factory(name, value, VrfRtImportMap)

    def _create_import_map_factory(self, name, value, cls):
        if self._import_maps_by_name.get(name) is not None:
            raise ImportMapAlreadyExistsError()
        self._import_maps_by_name[name] = cls(value)

    def get_import_map_by_name(self, name):
        return self._import_maps_by_name.get(name)


class ImportMapAlreadyExistsError(Exception):
    pass



# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

NETFLOW_V1 = 0x01
NETFLOW_V5 = 0x05
NETFLOW_V6 = 0x06
NETFLOW_V7 = 0x07
NETFLOW_V8 = 0x08
NETFLOW_V9 = 0x09


class NetFlow(object):
    _PACK_STR = '!H'
    _NETFLOW_VERSIONS = {}

    @staticmethod
    def register_netflow_version(version):
        def _register_netflow_version(cls):
            NetFlow._NETFLOW_VERSIONS[version] = cls
            return cls
        return _register_netflow_version

    def __init__(self):
        super(NetFlow, self).__init__()

    @classmethod
    def parser(cls, buf):
        (version,) = struct.unpack_from(cls._PACK_STR, buf)

        cls_ = cls._NETFLOW_VERSIONS.get(version, None)
        if cls_:
            return cls_.parser(buf)
        else:
            return None


@NetFlow.register_netflow_version(NETFLOW_V5)
class NetFlowV5(object):
    _PACK_STR = '!HHIIIIBBH'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, version, count, sys_uptime, unix_secs,
                 unix_nsecs, flow_sequence, engine_type, engine_id,
                 sampling_interval, flows=None):
        self.version = version
        self.count = count
        self.sys_uptime = sys_uptime
        self.unix_secs = unix_secs
        self.unix_nsecs = unix_nsecs
        self.flow_sequence = flow_sequence
        self.engine_type = engine_type
        self.engine_id = engine_id
        self.sampling_interval = sampling_interval

    @classmethod
    def parser(cls, buf):
        (version, count, sys_uptime, unix_secs, unix_nsecs,
         flow_sequence, engine_type, engine_id, sampling_interval) = \
            struct.unpack_from(cls._PACK_STR, buf)

        msg = cls(version, count, sys_uptime, unix_secs, unix_nsecs,
                  flow_sequence, engine_type, engine_id,
                  sampling_interval)
        offset = cls._MIN_LEN
        msg.flows = []
        while len(buf) > offset:
            f = NetFlowV5Flow.parser(buf, offset)
            offset += NetFlowV5Flow._MIN_LEN
            msg.flows.append(f)

        return msg


class NetFlowV5Flow(object):
    _PACK_STR = '!IIIHHIIIIHHxBBBHHBB2x'
    _MIN_LEN = struct.calcsize(_PACK_STR)

    def __init__(self, srcaddr, dstaddr, nexthop, input_, output,
                 dpkts, doctets, first, last, srcport, dstport,
                 tcp_flags, prot, tos, src_as, dst_as, src_mask,
                 dst_mask):
        self.srcaddr = srcaddr
        self.dstaddr = dstaddr
        self.nexthop = nexthop
        self.input = input_
        self.output = output
        self.dpkts = dpkts
        self.doctets = doctets
        self.first = first
        self.last = last
        self.srcport = srcport
        self.dstport = dstport
        self.tcp_flags = tcp_flags
        self.prot = prot
        self.tos = tos
        self.src_as = src_as
        self.dst_as = dst_as
        self.src_mask = src_mask
        self.dst_mask = dst_mask

    @classmethod
    def parser(cls, buf, offset):
        (srcaddr, dstaddr, nexthop, input_, output, dpkts, doctets,
         first, last, srcport, dstport, tcp_flags, prot, tos, src_as,
         dst_as, src_mask, dst_mask) = struct.unpack_from(
            cls._PACK_STR, buf, offset)
        msg = cls(srcaddr, dstaddr, nexthop, input_, output, dpkts,
                  doctets, first, last, srcport, dstport, tcp_flags,
                  prot, tos, src_as, dst_as, src_mask, dst_mask)

        return msg
